

# OGR driver names
OGR_FORMAT_ESRI_SHAPEFILE = "ESRI SHAPEFILE"
OGR_FORMAT_POSTGRESQL = "POSTGRESQL"
OGR_FORMAT_KML = "KML"
OGR_FORMAT_XLSX = "XLSX"
OGR_FORMAT_CSV = "CSV"
OGR_FORMAT_GEOJSON = "GEOJSON"
OGR_FORMAT_GML = "GML"
OGR_FORMAT_MAPINFO = "MAPINFO FILE"
OGR_DRIVER_NAMES = {OGR_FORMAT_ESRI_SHAPEFILE,
                    OGR_FORMAT_POSTGRESQL,
                    OGR_FORMAT_KML,
                    OGR_FORMAT_XLSX,
                    OGR_FORMAT_CSV,
                    OGR_FORMAT_GEOJSON,
                    OGR_FORMAT_GML,
                    OGR_FORMAT_MAPINFO
                    }


field_driver_uncompatibility = {
    (OGR_FORMAT_GEOJSON, 'Binary'): "OGR/GDAL does not allow to use binary field's type '{field_name}' in {ogr_format}."
                                    " You should convert it to string format."
}


import time
import inspect

from geoformat.conf.timer import Timer


def timeit(func):

    def timed(*args, **kw):
        ts = time.time()
        result = func(*args, **kw)
        te = time.time()
        execute_time = te - ts
        current_function_name = func.__name__

        if 'log_time' in kw:
            function_timer = kw['log_time']
            if function_timer:
                for stac in reversed(inspect.stack()):
                    if stac.function not in {'timed', '<module>'}:
                        if stac.function in function_timer.get_children_name():
                            function_timer = function_timer.get_child_timer(child_name=stac.function)
                        else:
                            # create new Timer
                            child_timer = Timer(timer_name=stac.function)
                            # put Timer in child
                            function_timer.add_child_timer(child_timer_object=child_timer)
                            function_timer = function_timer.get_child_timer(stac.function)

                if current_function_name in function_timer.function_in_it:
                    child_timer = function_timer.get_child_timer(current_function_name)
                    child_timer.add_time(execute_time=execute_time)
                else:
                    child_timer = Timer(timer_name=current_function_name)
                    child_timer.add_time(execute_time=execute_time)
                    # save child in parent timer
                    function_timer.add_child_timer(child_timer_object=child_timer)

                # compute decorator time
                decorator_time = time.time() - te
                child_timer.add_decorator_time(execute_time=decorator_time)



        else:
            print('{current_function_name} : {execute_time} ms'.format(
                    current_function_name=current_function_name,
                    execute_time=execute_time*1000
                )
            )

        return result

    return timed




import datetime

geoformat_field_type_to_python_type = {
    "Integer": int,
    "IntegerList": (int, list),
    "Real": float,
    "RealList": (float, list),
    "String": str,
    "StringList": (str, list),
    "Binary": bytes,
    "Date": datetime.date,
    "Time": datetime.time,
    "DateTime": datetime.datetime,
    "Boolean": bool,
}

geoformat_field_type_to_postgresql_type = {
    "Integer": "integer",
    "IntegerList": "integer[]",
    "Real": "numeric",
    "RealList": "numeric[]",
    "String": "character varying",
    "StringList": "character varying[]",
    "Binary": "bytea",
    "Date": "date",
    "Time": "time",
    "DateTime": "timestamp",
    "Boolean": "boolean",
}

geoformat_field_type_to_geojson_field_type = {
    "Binary": str,
    "Date": str,
    "Time": str,
    "DateTime": str,
}


shapefile_str_max_width = 250
geoformat_field_type_to_shapefile_field_type = {
    "Integer": "N",
    "IntegerList": None,
    "Real": "N",
    "RealList": None,
    "String": "C",
    "StringList": None,
    "Binary": None,
    "Date": "D",
    "Time": "C",
    "DateTime": "C",
    "Boolean": "L",
}

"""

    "C": Characters, text.
    "N": Numbers, with or without decimals.
    "F": Floats (same as "N").
    "L": Logical, for boolean True/False values.
    "D": Dates.
    "M": Memo, has no meaning within a GIS and is part of the xbase spec instead.

"""

shapefile_field_type_to_geoformat_field_type = {
    "C": "String",
    "N": "Real",
    "Float": "Real",
    "L": "Boolean",
    "D": "Date"
}

# all in str except String itself
geoformat_field_type_to_csv_field_type = {
    "Integer": str,
    "IntegerList": str,
    "Real": str,
    "RealList": str,
    "StringList": str,
    "Binary": str,
    "Date": str,
    "Time": str,
    "DateTime": str,
    "Boolean": str,
}

python_type_to_geoformat_field_type = {
    str: "String",
    (str, list): "StringList",
    float: "Real",
    (float, list): "RealList",
    int: "Integer",
    (int, list): "IntegerList",
    datetime.date: "Date",
    datetime.time: "Time",
    datetime.datetime: "DateTime",
    bytes: "Binary",
    bool: "Boolean",
}

recast_black_list = {
    "Integer": {"Binary"},
    "IntegerList": {"Real", "Integer", "Binary", "Date", "Time", "DateTime", "Boolean"},
    "Real": {"Binary"},
    "RealList": {"Integer", "Real", "Binary", "Date", "Time", "DateTime", "Boolean"},
    "String": {"Date", "Time", "DateTime"},
    "StringList": {"Integer", "Real", "Binary", "Date", "Time", "DateTime", "Boolean"},
    "Date": {
        "Binary",
        "Time",
        "DateTime",
    },
    "Time": {
        "Binary",
        "Date",
        "DateTime",
    },
    "DateTime": {
        "Binary"
    },
    "Binary": {
        "Integer",
        "IntegerList",
        "Real",
        "RealList",
        "Date",
        "Time",
        "DateTime",
    },
    "Boolean": {"IntegerList", "RealList", "StringList", "Date", "Time", "DateTime"},
}
ogr_field_type_to_geoformat_field_type = {
    0: "Integer",
    1: "IntegerList",
    2: "Real",
    3: "RealList",
    4: "String",
    5: "StringList",
    6: "String",  # WideString
    7: "StringList",  # WideStringList
    8: "Binary",
    9: "Date",
    10: "Time",
    11: "DateTime",
}

field_type_set = {
    "Integer",
    "IntegerList",
    "Real",
    "RealList",
    "String",
    "StringList",
    "Binary",
    "Date",
    "Time",
    "DateTime",
    "Boolean",
}
field_metadata_width_required = {"Real", "RealList", "String", "StringList"}
field_metadata_precision_required = {"Real", "RealList"}
none_value_pattern = {None}



from geoformat.conversion.geolayer_conversion import create_geolayer_from_i_feat_list
from geoformat.conf.error_messages import field_missing


####
#
# CLAUSE
#
# The CLAUSE functions return a directly a list of i_feat or a dict with a list associates to each key of i_feat
###

def clause_where(geolayer, field_name, predicate, values):
    """
    This function return an i_feat list from a geolayer that match with a predicate and given value.

    :param geolayer: layer on which we want to apply the where clause
    :param field_name: name of field on which is made the where clause
    :param predicate: predicate used in this list : '=','<', '>', '<>', 'LIKE', 'BETWEEN', 'IS', 'IS NOT'
    :param values: value that we want match or unmatch

    :return: i_feat_list
    """
    if field_name not in geolayer['metadata']['fields']:
        raise Exception(field_missing.format(field_name=field_name))

    # transform values to iterate variable
    if not isinstance(values, (list, tuple)):
        if isinstance(values, tuple):
            values = list(values)
        else:
            values = [values]

    # init result list
    i_feat_list = []

    # loop on geolayer
    for i_feat in geolayer['features']:
        feature = geolayer['features'][i_feat]
        feature_value = feature.get('attributes', {}).get(field_name)

        # save feature_value in list by default
        if isinstance(feature_value, (list, tuple, set)):
            feature_value_list = feature_value
        else:
            feature_value_list = [feature_value]

        save_i_feat = False
        # loop on feature_value_list
        for feature_value in feature_value_list:

            if '=' in predicate or 'IN' in predicate or predicate == 'LIKE' or predicate == 'BETWEEN':
                if feature_value in values:
                    save_i_feat = True
                    break

            if '>' in predicate or '<' in predicate or predicate == 'BETWEEN':
                if predicate == '<>':
                    if feature_value not in values:
                        save_i_feat = True
                else:
                    if feature_value is not None:
                        if '>' in predicate and feature_value > values[0]:
                            save_i_feat = True
                        if '<' in predicate and feature_value < values[-1]:
                            save_i_feat = True
                        if predicate == 'BETWEEN':
                            if values[0] < feature_value < values[-1]:
                                save_i_feat = True

            if 'IS' in predicate:
                if feature_value in values:
                    save_i_feat = True

                if 'NOT' in predicate:
                    save_i_feat = not save_i_feat

        if save_i_feat:
            i_feat_list.append(i_feat)

    return i_feat_list


def clause_group_by(geolayer, field_name_list):
    """
    Return a dictionnary with field name list as key and i_feat list from geolayer
    """

    if isinstance(field_name_list, str):
        field_name_list = [field_name_list]

    result_dico = {}
    for i_feat in geolayer['features']:
        feature = geolayer['features'][i_feat]

        # if feature is serialized
        if 'feature_serialize' in geolayer['metadata']:
            if geolayer['metadata']['feature_serialize']:
                feature = eval(feature)

        field_value_tuple = tuple(
            [feature['attributes'][field_name] if field_name in feature['attributes'] else None for field_name in
             field_name_list])

        # convert list value to tuple (if exists) very rare
        field_value_tuple = tuple([tuple(value) if isinstance(value, list) else value for value in field_value_tuple])

        if field_value_tuple in result_dico:
            result_dico[field_value_tuple].append(i_feat)
        else:
            result_dico[field_value_tuple] = [i_feat]

    return result_dico


def clause_order_by(geolayer, order_parameters):
    """
    Send i_feat list in order define in order_parameters

    order_parameters format : 3 poosibilities :
        order_parameters = 'field_name' (order by default is ASC)
        order_parameters = ('field_name', 'ASC' or 'DESC')
        order_parameter = [('field_name_a', 'ASC' or 'DESC'), ('field_name_b', 'ASC' or 'DESC'), ..., ('field_name_n', 'ASC' or 'DESC'
    """

    def insert_in_order_list_by_split(value, i_feat_list, order_by_list):
        # try to insert at extremity of list
        # value is lower of last value in order_by_list
        if value <= order_by_list[0][0]:
            if value == order_by_list[0][0]:
                order_by_list[0][1] = order_by_list[0][0] + i_feat_list
            else:
                order_by_list = [(value, i_feat_list)] + order_by_list
            return order_by_list
        # value is upper of last value in order_by_list
        elif value >= order_by_list[-1][0]:
            if value == order_by_list[-1][0]:
                order_by_list[-1][1] = order_by_list[-1][0] + i_feat_list
            else:
                order_by_list = order_by_list + [(value, i_feat_list)]
            return order_by_list

        # if no insertion we have to split in two
        list_cuter_idx = int(len(order_by_list) / 2)
        if value <= order_by_list[list_cuter_idx - 1][0]:
            if value == order_by_list[list_cuter_idx - 1][0]:
                order_by_list[list_cuter_idx - 1][1] += i_feat_list
                return order_by_list
            else:
                order_by_list_splited = order_by_list[:list_cuter_idx]
                result = insert_in_order_list_by_split(value, i_feat_list, order_by_list_splited)
                return result + order_by_list[list_cuter_idx:]

        if value >= order_by_list[list_cuter_idx][0]:
            if value == order_by_list[list_cuter_idx][0]:
                order_by_list[list_cuter_idx][1] += i_feat_list
                return order_by_list
            else:
                order_by_list_splited = order_by_list[list_cuter_idx:]
                result = insert_in_order_list_by_split(value, i_feat_list, order_by_list_splited)
                return order_by_list[:list_cuter_idx] + result

        return order_by_list[:list_cuter_idx] + [(value, i_feat_list)] + order_by_list[list_cuter_idx:]

    def order_values(dico_value, field_order):
        """
            This function ordered value in function of field order
        """
        order_by_list = None
        none_value_i_feat = []
        for value in dico_value:
            i_feat_list = dico_value[value]
            if None in value:
                none_value_i_feat += i_feat_list
            else:
                # first iteration
                if not order_by_list:
                    order_by_list = [(value, i_feat_list)]
                else:
                    order_by_list = insert_in_order_list_by_split(value, i_feat_list, order_by_list)

        # reverse order if we are DESC
        if field_order.upper() == 'DESC' and order_by_list:
            order_by_list = sorted(order_by_list, reverse=True)

        # add none value (if exists) in function of order_fields
        if len(none_value_i_feat) > 0:
            if order_by_list:
                if field_order.upper() == 'ASC':
                    order_by_list += [((None,), none_value_i_feat)]
                else:
                    order_by_list = [((None,), none_value_i_feat)] + order_by_list
            else:
                order_by_list = [(None, none_value_i_feat)]

        return order_by_list

    def field_group_by_then_order(geolayer, order_parameters):
        """
        This function is used to realise first a group by and reorder result
        Then it recall it if there is an other field_order_paramenters
        """
        (field_name, field_order) = order_parameters[0]
        gb_field_name = clause_group_by(geolayer, field_name)
        gb_field_name_ordered = order_values(gb_field_name, field_order)
        result_i_feat_list = []
        for value, i_feat_list in gb_field_name_ordered:
            if len(i_feat_list) > 1:
                if len(order_parameters) > 1:
                    new_order_parameters = order_parameters[1:]
                    new_geolayer = create_geolayer_from_i_feat_list(geolayer, i_feat_list, reset_i_feat=False)
                    result_i_feat_list += field_group_by_then_order(new_geolayer, new_order_parameters)
                else:
                    result_i_feat_list += i_feat_list
            else:
                result_i_feat_list += i_feat_list

        return result_i_feat_list

    # verification enters parameters
    if isinstance(order_parameters, (list, tuple)):
        authorised_order_value = set(['ASC', 'DESC'])
        for i_field, field in enumerate(order_parameters):
            if len(field) == 1:
                order_parameters[i_field] = (field, 'ASC')
            elif len(field) == 2:
                if field[1].upper() not in authorised_order_value:
                    raise Exception('error you must add "ASC" or "DESC" key in second position')
            else:
                raise Exception('error order_paramaters variable format is not correct')
    elif isinstance(order_parameters, str):
        order_parameters = [(order_parameters, 'ASC')]
    else:
        raise Exception("error: order_parameters must be a list / tuple or str value")

    for (field_name, order_value) in order_parameters:
        if field_name not in geolayer['metadata']['fields']:
            raise Exception(field_missing.format(field_name=field_name))

    return field_group_by_then_order(geolayer, order_parameters)




from geoformat.conversion.geometry_conversion import (
    geometry_to_ogr_geometry,
    ogr_geometry_to_geometry
)


def union_by_split(geometry_list, split_factor=2, bbox=True):
    """
    Union geometry list with split list method (split_factor default 2 by 2)
    * OGR dependencie : Union *
    """

    len_list = len(geometry_list)
    if len_list > split_factor:
        union_geom_list = []
        for i in range(split_factor):
            # first iteration
            if i == 0:
                split_geometry_list = geometry_list[:int(len_list / split_factor)]
            # last iteration
            elif i == split_factor - 1:
                begin = int(len_list / split_factor) * i
                split_geometry_list = geometry_list[begin:]
            # others iterations
            else:
                begin = int(len_list / split_factor) * i
                end = (int(len_list / split_factor)) * (i + 1)
                split_geometry_list = geometry_list[begin:end]

            # adding union to geom list
            union_geom_list.append(union_by_split(split_geometry_list, split_factor))

        for i, union_geom in enumerate(union_geom_list):
            if i == 0:
                ogr_geom_unioned = geometry_to_ogr_geometry(union_geom)
            else:
                temp_geom = geometry_to_ogr_geometry(union_geom)
                ogr_geom_unioned = ogr_geom_unioned.Union(temp_geom)

        geojson_unioned = ogr_geometry_to_geometry(ogr_geom_unioned, bbox=bbox)

        return geojson_unioned

    else:
        if len(geometry_list) > 1:

            for i, union_geom in enumerate(geometry_list):
                if i == 0:
                    ogr_geom_unioned = geometry_to_ogr_geometry(union_geom)
                else:
                    temp_geom = geometry_to_ogr_geometry(union_geom)
                    ogr_geom_unioned = ogr_geom_unioned.Union(temp_geom)

            geojson_unioned = ogr_geometry_to_geometry(ogr_geom_unioned, bbox=bbox)

            return geojson_unioned

        else:
            return geometry_list[0]



from geoformat.conf.fields_variable import geoformat_field_type_to_postgresql_type
from geoformat.conversion.feature_conversion import feature_deserialize
from geoformat.conversion.geometry_conversion import geometry_to_wkb, geometry_to_wkt

try:
    import psycopg2
    import psycopg2.extras
    import_psycopg2_success = True
except ImportError:
    import_psycopg2_success = False

from geoformat.conf.error_messages import import_psycopg2_error


def fields_metadata_to_create_table_fields_structure(
        geolayer_fields_metadata,
        consider_width_and_precision
):
    """
    This function convert geolayer's fields metadata to

    :param geolayer_fields_metadata: geolayer's fields metadata.
    :param consider_width_and_precision: True if you want consider field with and precision / False if you won't.
    :return:
    """
    field_structure_list = [None] * len(geolayer_fields_metadata)

    for i_field, (field_name, field_metadata) in enumerate(geolayer_fields_metadata.items()):
        field_type = field_metadata['type']
        field_type_pg = geoformat_field_type_to_postgresql_type[field_type]
        if consider_width_and_precision:
            if field_type in {'Real', 'String'}:
                width = field_metadata['width']
                if field_type == 'Real':
                    precision = field_metadata['precision']
                    suffix = '({width}, {precision})'.format(width=width, precision=precision)
                else:
                    if width > 0:
                        suffix = '({width})'.format(width=width)
                    else:
                        suffix = ''
                field_type_pg = field_type_pg + suffix

        field_structure = '{field_name} {field_type_pg}'.format(field_name=field_name, field_type_pg=field_type_pg)
        field_index = field_metadata.get('index', i_field)
        field_structure_list[field_index] = field_structure

    table_fields_structure = ', '.join(field_structure_list)

    return table_fields_structure


def geometry_ref_to_geometry_field_structure(
        geometry_ref,
        geometry_field_name,
        constraint_geometry_type,
):
    srid = geometry_ref.get('crs', None)
    geometry_type = geometry_ref.get('type', None)
    geometry_field = '{geometry_field_name} geometry'.format(geometry_field_name=geometry_field_name)
    add_geometry_metadata = []

    # get geometry type
    if geometry_type:
        if constraint_geometry_type is True and len(geometry_type) == 1:
            geometry_type = list(geometry_type)[0]
            add_geometry_metadata.append("'{geometry_type}'".format(geometry_type=geometry_type))
        else:
            add_geometry_metadata.append("'Geometry'")
    # get srid
    if srid:
        add_geometry_metadata.append('{srid}'.format(srid=srid))

    # add geometry metadata
    if add_geometry_metadata:
        if len(add_geometry_metadata) > 1:
            add_geometry_metadata = ', '.join(add_geometry_metadata)
        else:
            add_geometry_metadata = add_geometry_metadata[0]

        geometry_field = '{geometry_field}({add_geometry_metadata})'.format(
            geometry_field=geometry_field,
            add_geometry_metadata=add_geometry_metadata
        )

    return geometry_field


def format_to_posgresq_values(geolayer):

    fields_metadata = geolayer['metadata'].get('fields', [])
    nb_features = len(geolayer['features'])
    nb_fields = len(fields_metadata)

    # add geometry field if necessary
    if 'geometry_ref' in geolayer['metadata']:
        nb_fields += 1
        srid = geolayer['metadata']['geometry_ref'].get('crs', None)

    field_structure_list = [None] * nb_fields
    # if there is field
    if fields_metadata:
        for i_field, (field_name, field_metadata) in enumerate(fields_metadata.items()):
            field_index = field_metadata.get('index', i_field)
            field_structure_list[field_index] = field_name

    # if features are serialized
    serialize = geolayer['metadata'].get('feature_serialize', False)

    values = [None] * nb_features
    for ii, (i_feat, feature) in enumerate(geolayer['features'].items()):
        feature_values_format = [None] * nb_fields
        # if features are serialized in geolayer we deserialize it
        if serialize is True:
            feature = feature_deserialize(feature, bbox=False)

        # add attributes data if exists
        feature_attributes = feature.get('attributes', None)
        if feature_attributes:
            # write field value if exists
            for i_field, field_name in enumerate(field_structure_list):
                if field_name in feature_attributes:
                    feature_values_format[i_field] = feature_attributes[field_name]

        # add geometry if exists
        if 'geometry' in feature:
            feature_geometry = feature['geometry']
            if srid:
                feature_geometry_wkt = geometry_to_wkt(feature_geometry)
                feature_geometry_for_pg = 'SRID={srid};{feature_geometry_wkt}'.format(
                    feature_geometry_wkt=feature_geometry_wkt,
                    srid=srid
                )
            else:
                feature_geometry_for_pg = geometry_to_wkb(feature_geometry)

            # add geometry to feature_values_format
            feature_values_format[-1] = feature_geometry_for_pg

        values[ii] = feature_values_format

    return values


def _method_execute_batch(cursor, schema, table_name, values):
    """
    Batch method to write data in postgresql table

    inspired by : https://www.datacareer.de/blog/improve-your-psycopg2-executions-for-postgresql-in-python/

    :param cursor: cursor
    :param schema:
    :param table_name:
    :param values:
    :return:
    """
    nb_fields = len(values[0])
    field_struct = ', '.join(['%s'] * nb_fields)
    insert_into_request = "INSERT INTO {schema}.{table} VALUES ({field_struct})".format(
        schema=schema,
        table=table_name,
        field_struct=field_struct
    )
    psycopg2.extras.execute_batch(
        cursor,
        insert_into_request,
        values
    )


def geolayer_to_postgres(
    geolayer,
    host,
    database_name,
    user,
    password,
    port=5432,
    schema='public',
    geometry_column_name='geom',
    overwrite=True,
    constraint_geometry_type=False,
    consider_width_and_precision=True
):
    """
    Convert geolayer to table in postgresql database.

    :param geolayer: input geolayer that we want create in postgresql
    :param host: host of postgresql server
    :param database_name: name of database
    :param user: name of user
    :param password: password for user
    :param port: port of database (default 5432).
    :param schema: name of the schema in which we want to insert the data (default public).
    :param geometry_column_name: name of geometry column (default geom).
    :param overwrite: True if table exists we drop it / False if table exists process fail (default True).
    :param constraint_geometry_type: geometry field in postgis can be forced to a single geometry type. True if want to
    add this constraint (it is only possible if geolayer contain only one geometry's type) / False if you wan't it
    (by default).
    :param consider_width_and_precision: add precision and width when create field (default True).
    """
    if import_psycopg2_success is True:
        # connect to database
        conn_string = "host={host} user={user} dbname={dbname} password={password} port={port}".format(
            host=host,
            user=user,
            password=password,
            dbname=database_name,
            port=port)
        database_connection = psycopg2.connect(conn_string)
        cursor = database_connection.cursor()

        # get table name
        table_name = geolayer['metadata']['name']

        # create schema if not exists
        create_schema_request = """
        CREATE SCHEMA IF NOT EXISTS {schema};
        """.format(schema=schema)
        cursor.execute(create_schema_request)

        # if overwrite
        if overwrite:
            overwrite_request = """
            DROP TABLE IF EXISTS {schema}.{table_name};
            """.format(schema=schema, table_name=table_name)
            cursor.execute(overwrite_request)
            database_connection.commit()

        # get attributes field table structure
        geolayer_fields_metadata = geolayer['metadata'].get('fields', None)
        table_fields_structure = ''
        if geolayer_fields_metadata:
            table_fields_structure = fields_metadata_to_create_table_fields_structure(
                geolayer_fields_metadata=geolayer_fields_metadata,
                consider_width_and_precision=consider_width_and_precision
            )

        # add geometry field
        geolayer_geometry_metadata = geolayer['metadata'].get('geometry_ref', None)
        if geolayer_geometry_metadata:
            if table_fields_structure:
                table_fields_structure += ', '

            table_fields_structure += geometry_ref_to_geometry_field_structure(
                geolayer_geometry_metadata,
                geometry_column_name,
                constraint_geometry_type,
            )

        # create table
        create_table_request = """
        CREATE TABLE {schema}.{table_name} ({table_fields_structure});
        """.format(schema=schema, table_name=table_name, table_fields_structure=table_fields_structure)
        cursor.execute(create_table_request)
        database_connection.commit()

        # insert into with batch execute
        # format data
        values = format_to_posgresq_values(geolayer)

        # insert into db
        _method_execute_batch(
            cursor=cursor,
            schema=schema,
            table_name=table_name,
            values=values
        )

        # end of transaction
        database_connection.commit()
    else:
        raise Exception(import_psycopg2_error)


import tempfile
from pathlib import Path

from geoformat.conf.path import (
    verify_input_path_is_file,
    verify_input_path_is_http,
    path_is_file,
    path_is_http,
    open_http_path,
)

from geoformat.conf.error_messages import path_not_exists

def _get_recast_field_type_mapping(fields_metadata, geoformat_type_to_output_driver_type):
    """
    From geolayer field's metadata to output driver recast dict

    :param fields_metadata: geolayer field's metadata
    :param geoformat_type_to_output_driver_type: dict that make translation between geoformat type and output driver
    type

    :return: dict that allow to recast field
    """
    recast_field_mapping = {}
    if fields_metadata:
        for field_name, field_metadata in fields_metadata.items():
            recast_to_python_type = geoformat_type_to_output_driver_type.get(field_metadata['type'], None)
            if recast_to_python_type is not None:
                recast_field_mapping[field_name] = {
                    "recast_value_to_python_type": recast_to_python_type,
                    "resize_value_width": None,
                    "resize_value_precision": None
                }

    return recast_field_mapping


def load_data(path, encoding='utf8', http_headers=None, read_mode='r',  temp_mode='w',):
    """
    From path return data as a NamedTemporaryFile object

    :param path: file path or http path
    :param encoding: specify file char encoding (default value utf8)
    :param http_headers: optionally you can add headers of http parameters
    :param read_mode: data is open in 'r' mode (by default) but you can choose an other parameters ('rb', 'r+b')
    :param temp_mode: data is load in NamedTemporaryFile we can choose writing mode
    :return:
    """
    # create temporary file
    fp = tempfile.NamedTemporaryFile(mode=temp_mode)

    if path_is_file(path=path) is True:
        p = verify_input_path_is_file(path=path)
        # open file
        with open(file=p, mode=read_mode, encoding=encoding) as file:
            fp.write(file.read())

    elif path_is_http(path=path, headers=http_headers)[0] is True:
        p = verify_input_path_is_http(path=path, headers=http_headers)
        http_req = open_http_path(path=p, headers=http_headers)
        http_req_str = http_req.read()
        if encoding is not None:
            http_req_str = http_req_str.decode(encoding)
        fp.write(http_req_str)
        file_name = http_req.info().get_filename()
        if file_name is None:
            file_name = path.split('/')[-1]
        p = Path(file_name)
    else:
        raise Exception(path_not_exists)

    fp.seek(0)

    file_name = p.stem
    return fp, file_name



import copy

from geoformat import features_fields_type_scan, fields_scan_to_fields_metadata, features_geometry_ref_scan, \
    geometries_scan_to_geometries_metadata, feature_filter, feature_serialize, recast_field
from geoformat.conf.fields_variable import none_value_pattern
from geoformat.conf.geometry_variable import GEOFORMAT_GEOMETRY_TYPE
from geoformat.conf.error_messages import (
    field_name_not_indexing,
    key_use_only_by_geometry,
    geometry_not_in_variable,
    feature_missing
)
from geoformat.conf.format_data import value_to_iterable_value
from geoformat.conversion.fields_conversion import recast_field
from geoformat.conversion.feature_conversion import (
    feature_deserialize,
    feature_serialize,
)
from geoformat.geoprocessing.geoparameters.bbox import bbox_union
from geoformat.manipulation.feature_manipulation import (
    rename_field_in_feature,
    drop_field_in_feature,
)
from geoformat.manipulation.metadata_manipulation import (
    drop_field_in_metadata,
    rename_field_in_metadata,
    create_field_in_metadata,
    check_if_field_exists_in_metadata,
    check_attributes_index_in_metadata,
    add_attributes_index_in_metadata,
)


def delete_feature(geolayer, i_feat_to_delete):
    """Delete given i_feat feature from geolayer

    TODO add an update for geometry_ref metadata
    """
    i_feat_to_delete_list = value_to_iterable_value(
        value=i_feat_to_delete, output_iterable_type=list
    )
    for i_feat in i_feat_to_delete_list:
        if i_feat in geolayer["features"]:
            # TODO add delete (update) index (attributes and geometry) / matrix
            del geolayer["features"][i_feat]
        else:
            raise Exception(feature_missing.format(i_feat=i_feat))

    if not geolayer["features"]:
        if "fields" in geolayer["metadata"]:
            del geolayer["metadata"]["fields"]

        if "geometry_ref" in geolayer["metadata"]:
            del geolayer["metadata"]["geometry_ref"]

        if "index" in geolayer["metadata"]:
            del geolayer["metadata"]["index"]

        if "matrix" in geolayer["metadata"]:
            del geolayer["metadata"]["matrix"]

    return geolayer


def drop_field(geolayer, field_name_to_drop):
    """
    This function allow to drop a field in geolayer.

    :param geolayer: input geolayer.
    :param field_name_to_drop: field name in geolayer to drop. can be a fields list
    :return: geolayer with field drop.
    """
    field_name_to_drop = value_to_iterable_value(
        value=field_name_to_drop, output_iterable_type=list
    )
    geolayer["metadata"] = drop_field_in_metadata(
        metadata=geolayer["metadata"], field_name_or_field_name_list=field_name_to_drop
    )

    # delete field in features
    delete_i_feat_list = []
    for i_feat, feature in geolayer["features"].items():
        feature = drop_field_in_feature(
            feature=feature, field_name_or_field_name_list=field_name_to_drop
        )
        geolayer["features"][i_feat] = feature

        if not feature:
            delete_i_feat_list.append(i_feat)

    # delete feature with no data in it
    if delete_i_feat_list:
        geolayer = delete_feature(
            geolayer=geolayer, i_feat_to_delete=delete_i_feat_list
        )

    return geolayer


def rename_field(geolayer, old_field_name, new_field_name):
    """
    Rename field in geolayer

    :param geolayer: input geolayer.
    :param old_field_name: actual name of field that we want to change.
    :param new_field_name: new field's name that we want to apply.
    :return: geolayer with field rename
    """
    # rename field in metadata
    geolayer["metadata"] = rename_field_in_metadata(
        metadata=geolayer["metadata"],
        old_field_name=old_field_name,
        new_field_name=new_field_name,
    )
    # rename in features
    for i_feat, feature in geolayer["features"].items():
        feature = rename_field_in_feature(
            feature=feature,
            old_field_name=old_field_name,
            new_field_name=new_field_name,
        )
        geolayer["features"][i_feat] = feature

    return geolayer


def create_field(
    geolayer,
    field_name,
    field_type,
    field_width=None,
    field_precision=None,
    field_index=None,
):
    """
    Create new field in geolayer

    :param geolayer: input geolayer
    :param field_name: new field name
    :param field_type: field type
    :param field_width: field width
    :param field_precision: field precision
    :param field_index: field index
    :return: geolayer with new field in it
    """

    output_geolayer = copy.deepcopy(geolayer)

    new_metadata = create_field_in_metadata(
        metadata=geolayer["metadata"],
        field_name=field_name,
        field_type=field_type,
        field_width=field_width,
        field_precision=field_precision,
    )

    output_geolayer["metadata"] = new_metadata
    if new_metadata["fields"][field_name]["index"] != field_index:
        output_geolayer = recast_field(
            geolayer_to_recast=output_geolayer,
            field_name_to_recast=field_name,
            reindex=field_index,
        )

    return output_geolayer


def add_attributes_index(geolayer, field_name, index):
    """Store index in geolayer

    :param geolayer: geolayer to add an attributes
    :field_name: name of field concern by index
    :index: index that we want to add to field of geolayer
    :return: geolayer with index in it
    """

    metadata = add_attributes_index_in_metadata(
        metadata=geolayer["metadata"], field_name=field_name, index=index
    )
    geolayer["metadata"] = metadata

    return geolayer


def check_attributes_index(geolayer, field_name, type=None):
    """
    Return True or False if an attributes index in found in geolayer for input field_name optionaly we can test
    type index

    :param geolayer: input geolayer
    :param field_name: field name where is the index
    :param type: type of indexing (hashtable/btree ...)
    :return: Boolean
    """

    return check_attributes_index_in_metadata(
        metadata=geolayer["metadata"], field_name=field_name, type=type
    )


def delete_attributes_index(geolayer, field_name, type=None):
    """Delete attributes index when existing in geolayer. Opionnaly you can filter on index type

    :param geolayer: input Geolayer
    :param field_name: field name where is the index
    :param type: type of indexing (hashtable/btree ...)
    :return: Geolayer without index for given field_name (and optionally index type)
    """

    if check_attributes_index(geolayer, field_name, type) is True:
        del geolayer["metadata"]["index"]["attributes"][field_name]
        if len(geolayer["metadata"]["index"]["attributes"]) == 0:
            del geolayer["metadata"]["index"]["attributes"]
        if len(geolayer["metadata"]["index"]) == 0:
            del geolayer["metadata"]["index"]
    else:
        raise Exception(field_name_not_indexing.format(field_name=field_name))

    return geolayer


def check_if_field_exists(geolayer, field_name):
    """
    Check if field exists in given Geolayer.

    :param geolayer: input geolayer where existing field name is tested.
    :param field_name: name of field.
    :return: True if field exists False if not.
    """

    return check_if_field_exists_in_metadata(
        metadata=geolayer["metadata"], field_name=field_name
    )


def _get_driver_geometry_type_matching_i_feat_and_geometry_type_filter(
    geolayer,
    geometry_type_mapping,
):
    """
    Return a dict by geometry type fill in geometry_type_mapping variable.
    For each key is associate a list that contains dicts with features geolayer variable information in it.
        Each dict in list have this key value informations :
            i_feat : geolayer feature id
            i_geom : position of geometry in feature if SPLIT_BY_GEOMETRY_TYPE
            feature_geometry : output geometry

    SPLIT_BY_GEOMETRY_TYPE will divide a collection of geometry into geometries and distribute the geometries taking
    into account the characteristics set for each geometry in the variable: geometry_type_mapping.
    In case of EMPTY geometry collection an empty geometry is return for each type of geometry.

    TODO Usefully options that can be add further (if necessary)
    TODO SIMPLE_GEOMETRY
    TODO CONVERT_TO_POINT
    TODO CONVERT_TO_LINESTRING

    :param geolayer: input geolayer
    :param geometry_type_mapping: geometry mapping
    :return: dict described above
    """
    driver_i_feat_geometry_to_filter = {}
    geolayer_serialize = geolayer['metadata'].get("feature_serialize")
    # loop on each features
    for i_feat, feature in geolayer["features"].items():
        if geolayer_serialize is True:
            feature = feature_deserialize(feature, bbox=False)
        if feature.get("geometry"):
            geometry = feature["geometry"]
            geometry_type = geometry["type"]
            group_geometry_key = geometry_type_mapping.get(geometry_type)
            if group_geometry_key is None:
                raise Exception(geometry_not_in_variable.format(geometry_type=geometry_type, variable_name='geometry_type_mapping'))
            # put geometry(ies) in a list
            if group_geometry_key == "SPLIT_BY_GEOMETRY_TYPE":
                if 'geometries' in geometry:
                    geometry_list = geometry["geometries"]
                    if not geometry_list:  # if GEOMETRYCOLLECTION EMPTY
                        # create list with output geometry type
                        output_geometry_type_set = set()
                        output_geometry_type_list = []
                        for input_geometry, output_geometry in geometry_type_mapping.items():
                            if output_geometry not in output_geometry_type_set and output_geometry.upper() in GEOFORMAT_GEOMETRY_TYPE:
                                output_geometry_type_set.update([output_geometry])
                                output_geometry_type_list.append(output_geometry)

                        geometry_list = [{'type': geometry_type, "coordinates": []} for geometry_type in output_geometry_type_list]
                else:
                    raise Exception(key_use_only_by_geometry.format(
                        key="SPLIT_BY_GEOMETRY_TYPE",
                        geometry_type='GEOMETRYCOLLECTION'
                    ))
            else:
                geometry_list = [geometry]

            # loop on geometries
            if geometry_list:
                for i_geom, geometry_in_list in enumerate(geometry_list):
                    geometry_in_list_type = geometry_in_list["type"]
                    group_geometry_key_in_list = geometry_type_mapping[
                        geometry_in_list_type
                    ]

                    i_feat_matching_geom = {
                        "i_feat": i_feat,
                        "i_geom": i_geom,
                        "feature_geometry": geometry_in_list,
                    }
                    if driver_i_feat_geometry_to_filter.get(group_geometry_key_in_list):
                        driver_i_feat_geometry_to_filter[
                            group_geometry_key_in_list
                        ].append(i_feat_matching_geom)
                    else:
                        driver_i_feat_geometry_to_filter[group_geometry_key_in_list] = [
                            i_feat_matching_geom
                        ]
            # if no geometries
            else:
                pass

    return driver_i_feat_geometry_to_filter


def split_geolayer_by_geometry_type(
    geolayer,
    geometry_type_mapping
):
    """
    Split input geolayer according to the geometry_type_mapping variable.

    geometry_type_mapping is a dictionary witch key represents a geoformat geometry type and value represents the type
    of geometry we want to output.
    For GeometryCollection geometry type, only, you can use value SPLIT_BY_GEOMETRY_TYPE. This means that if a feature
    of the input geolayer contains a GeometryCollection in output this entity will be split according to the geometries
    contained in the GeometryCollection.

    Example of geometry_type_mapping :
    {
            "Point": "Point",
            "LineString": "LineString",
            "Polygon": "Polygon",
            "MultiPoint": "MultiPoint",
            "MultiLineString": "LineString",
            "MultiPolygon": "Polygon",
            "GeometryCollection": "SPLIT_BY_GEOMETRY_TYPE",
        }
        Here LineString and MultiLinestring, Polygon and Multipolygon will be in same output geolayer, Point and
        Multipoint geometries in separate Geolayer and entities GeometryCollection will this entity will be duplicated
        according to the geometries it contains.

    :param geolayer: geolayer that we want to split
    :param geometry_type_mapping:
    :yield: split geolayer
    """
    driver_i_feat_geometry_to_filter = _get_driver_geometry_type_matching_i_feat_and_geometry_type_filter(
        geolayer=geolayer,
        geometry_type_mapping=geometry_type_mapping
    )
    # if not split
    if len(driver_i_feat_geometry_to_filter) == 1:
        yield geolayer

    # we create a geolayer by geometry split
    else:
        for (
            driver_geometry_type,
            geolayer_i_feat_and_i_geom,
        ) in driver_i_feat_geometry_to_filter.items():
            # create geolayer
            output_geolayer_metadata = copy.deepcopy(geolayer["metadata"])
            output_geolayer_metadata["geometry_ref"]["type"] = {driver_geometry_type}
            extent_exists = output_geolayer_metadata["geometry_ref"].get("extent")
            if extent_exists:
                del output_geolayer_metadata["geometry_ref"]['extent']
                geolayer_extent = ()
            output_geolayer_metadata["name"] = (
                output_geolayer_metadata["name"] + "_" + driver_geometry_type
            )
            output_geolayer = {"metadata": output_geolayer_metadata, "features": {}}
            output_geolayer_serialize = output_geolayer['metadata'].get("feature_serialize")

            for i, i_feat_and_i_geom in enumerate(geolayer_i_feat_and_i_geom):
                # initialize feature
                feature = {}
                i_feat = i_feat_and_i_geom["i_feat"]
                feature_geometry = i_feat_and_i_geom["feature_geometry"]
                if feature_geometry:
                    if extent_exists:
                        feature_geometry_bbox = feature_geometry.get('bbox')
                        if feature_geometry_bbox:
                            geolayer_extent = bbox_union(bbox_a=geolayer_extent, bbox_b=feature_geometry_bbox)

                    feature["geometry"] = feature_geometry

                # get attributes
                feature_attributes = geolayer["features"][i_feat].get("attributes")
                if feature_attributes:
                    feature["attributes"] = feature_attributes

                if output_geolayer_serialize is True:
                    feature = feature_serialize(feature)
                # put feature in geolayer
                output_geolayer["features"][i] = feature

            # update extent
            if extent_exists:
                if geolayer_extent:
                    output_geolayer['metadata']['extent'] = geolayer_extent

            yield output_geolayer


def feature_list_to_geolayer(
        feature_list,
        geolayer_name,
        field_name_filter=None,
        force_field_conversion=False,
        geometry_type_filter=None,
        bbox_filter=None,
        bbox_extent=True,
        crs=None,
        serialize=False,
        none_value_pattern=none_value_pattern
):
    """
    Create a geolayer with an input feature list.

    :param feature_list: features list that we want to transform to geolayer.
    :param geolayer_name: name for geolayer.
    :param field_name_filter: field_name that we want to keep in geolayer (can be a list).
    :param force_field_conversion: True if you want to force value in field (can change field type) / False if you want
           to deduce field type without forcing field type.
    :param geometry_type_filter: keep only features with geometry type in this variable (can be a list).
    :param bbox_filter: keep only feature that intersects bbox (can be a list of bbox).
    :param bbox_extent: add "bbox" key in each features and "extent" key in geometry metadata.
    :param crs: coordinates spatial reference for geolayer.
    :param serialize: True if features in geolayer are serialized (can reduce performance) / False if not.
    :return: Return geolayer that contains features in feature_list depending on the options specified at the start
    of this function.
    """
    # initialize input variable
    feature_list = value_to_iterable_value(feature_list, list)
    field_name_filter = value_to_iterable_value(field_name_filter, list)
    geometry_type_filter = value_to_iterable_value(geometry_type_filter, list)
    if isinstance(bbox_filter, (list, tuple)):  # TODO remove when bbox will be an object
        if isinstance(bbox_filter[0], (int, float)):  # TODO remove when bbox will be an object
            bbox_filter = [bbox_filter]
    bbox_filter = value_to_iterable_value(bbox_filter, set)

    # initialize output variable
    # create empty _geolayer
    _geolayer = {
        "metadata": {
            "name": geolayer_name
        },
        "features": {

        }
    }
    # metadata
    # get fields metadata
    features_fields_scan = features_fields_type_scan(
        geolayer_or_feature_list=feature_list,
        field_name_filter=field_name_filter,
        try_to_force_type=force_field_conversion,
        fields_index=True,
        none_value_pattern=none_value_pattern
    )
    geolayer_fields_metadata = fields_scan_to_fields_metadata(
        fields_scan=features_fields_scan
    )
    # get geometry metadata
    features_geometries_scan = features_geometry_ref_scan(
        geolayer_or_feature_list=feature_list,
        geometry_type_filter=geometry_type_filter,
        bbox_filter=bbox_filter,
        extent=bbox_extent
    )
    geolayer_geometry_metadata = geometries_scan_to_geometries_metadata(
        geometry_scan=features_geometries_scan,
        crs=crs,
        extent=bbox_extent
    )

    # add metadata to _geolayer
    if geolayer_fields_metadata:
        _geolayer['metadata']['fields'] = geolayer_fields_metadata
    if geolayer_geometry_metadata:
        _geolayer['metadata']['geometry_ref'] = geolayer_geometry_metadata
    if serialize is True:
        _geolayer['metadata']['feature_serialize'] = True

    # add feature to _geolayer
    i_feat = 0

    for feature in feature_list:
        feature = feature_filter(
            feature=feature,
            field_name_filter=field_name_filter,
            geometry_type_filter=geometry_type_filter,
            bbox_filter=bbox_filter,
            bbox=bbox_extent
        )
        # if feature
        if feature:
            # if feature must be serialized
            if serialize:
                feature = feature_serialize(feature)
            _geolayer['features'][i_feat] = feature
            i_feat += 1

    # drop field if necessary
    if geolayer_fields_metadata:
        for field_name, field_dict in features_fields_scan.items():
            if field_dict['field_delete'] is True:
                _geolayer = drop_field(geolayer=_geolayer, field_name_to_drop=field_name)

    # recast field if necessary
    if geolayer_fields_metadata:
        for field_name, field_dict in features_fields_scan.items():
            if field_dict['field_recast'] is True and field_dict['field_delete'] is False:
                # define type to recast
                recast_to_type = _geolayer['metadata']['fields'][field_name]['type']
                # resize width to recast
                resize_width = None
                if 'width' in _geolayer['metadata']['fields'][field_name]:
                    resize_width = _geolayer['metadata']['fields'][field_name]['width']
                # resize precision to recast
                resize_precision = None
                if 'precision' in _geolayer['metadata']['fields'][field_name]:
                    resize_precision = _geolayer['metadata']['fields'][field_name]['precision']
                # recast field in _geolayer
                _geolayer = recast_field(
                    geolayer_to_recast=_geolayer,
                    field_name_to_recast=field_name,
                    recast_to_geoformat_type=recast_to_type,
                    resize_width=resize_width,
                    resize_precision=resize_precision,
                    none_value_pattern=none_value_pattern
                )

    return _geolayer



def get_slope_between_two_points(point_a, point_b):
    """
    Return slope value for a line that pass between two input points

    :param point_a: first input point
    :param point_b: second input point
    :return:  slope value
    """
    x_a, y_a = point_a
    x_b, y_b = point_b
    if x_a != x_b:
        slope = (y_b - y_a) / (x_b - x_a)
    else:
        slope = 'VERTICAL'

    return slope


def get_intercept_from_point_and_slope(point, slope):
    """
    return intercept value from a line that pass to input point with input slope

    :param point: point
    :param slope: slope value
    :return: intercept value that pass to input point with input slope
    """

    point_x, point_y = point

    # if slope is horizontal or vertical
    if slope == 0 or slope == 'VERTICAL':
        intercept = point_y
    else:
        intercept = slope * (0 - point_x) + point_y

    return intercept


def get_slope_from_point_and_intercept(point, intercept):
    """
    return slope value from a line that pass by input point and that intercepts y axis to input intercept value

    :param point:
    :param intercept:
    :return:
    """

    # if intercept is False slope is vertical
    if intercept is False:
        point_intercept = point
    else:
        point_intercept = 0, intercept

    return get_slope_between_two_points(point, point_intercept)


def line_parameters(segment):
    """
    Deduces from a segment lines (slope and  intercept) parameters passing through this segment

    :param segment: input segment
    :return: a dict with two linears parameters : slope AND y_intercept
    """

    ((x_a, y_a), (x_b, y_b)) = segment
    if x_b != x_a:
        a = (y_b - y_a) / (x_b - x_a)
        b = y_a - (x_a * a)
    else:
        a = 'VERTICAL'
        b = x_a

    return {'slope': a, 'intercept': b}


def perpendicular_line_parameters_at_point(line_parameters, point):
    """
    Gives perpendicular lines parameters of a line passing through a point from an original line parameters

    :param line_parameters: original line parameter
    :param point: point of origin which serves as a pivot for calculating the parameters of perpendicular lines
    :return:
    """

    a = line_parameters['slope']

    x_a, y_a = point
    if a == 0.:
        a_bis = 'VERTICAL'
        b_bis = x_a
    elif a == 'VERTICAL':
        a_bis = 0.
        b_bis = y_a
    else:
        a_bis = -1 / a
        b_bis = y_a - (x_a * a_bis)

    return {"slope": a_bis, 'intercept': b_bis}


def point_at_distance_with_line_parameters(start_point, line_parameters, distance, offset_distance=None):
    """
    Returns a point at distance to a input_point with a direction define by line parameters.
    Distance can be negative.
    optional: add an offset distance (perpendicular to line_parameters value)

    :param start_point: input start_point
    :param line_parameters: line parameter that define direction to the new start_point
    :param distance: distance between input start_point to output start_point
    :param offset_distance: in option you can add a perpendicular offset input start_point
    :return: output start_point
    """

    slope = line_parameters['slope']
    intercept = line_parameters['intercept']

    x_a, y_a = start_point
    if slope == 'VERTICAL':
        x_b = x_a
        y_b = y_a + distance
    else:
        x_b = x_a + (distance / ((1 + slope ** 2) ** 0.5))
        y_b = slope * x_b + intercept

    ouput_point = x_b, y_b

    if offset_distance:
        perpendicular_parameter = perpendicular_line_parameters_at_point(line_parameters=line_parameters,
                                                                         point=ouput_point)
        offset_x, offset_y = point_at_distance_with_line_parameters(start_point=ouput_point,
                                                                    line_parameters=perpendicular_parameter,
                                                                    distance=offset_distance,
                                                                    offset_distance=None)
        ouput_point = offset_x, offset_y

    return ouput_point


def crossing_point_from_lines_parameters(line_parameter_a, line_parameter_b):
    """
    Return point coordinates that crossing two lines (if possible: slope must be different).
    If lines are parallel the function returns None

    :param line_parameter_a: parameters for line a
    :param line_parameter_b: parameters for line b
    :return: intersecting point between two lines
    """
    # slope must be different
    if line_parameter_a['slope'] != line_parameter_b['slope']:
        # define x
        if line_parameter_a['slope'] == 'VERTICAL':
            point_x = line_parameter_a['intercept']
        elif line_parameter_b['slope'] == 'VERTICAL':
            point_x = line_parameter_b['intercept']
        else:
            point_x = (line_parameter_a['intercept'] - line_parameter_b['intercept']) / (line_parameter_b['slope'] - line_parameter_a['slope'])

        # define y
        if line_parameter_a['slope'] == 'VERTICAL':
            point_y = line_parameter_b['slope'] * point_x + line_parameter_b['intercept']
        else:
            point_y = line_parameter_a['slope'] * point_x + line_parameter_a['intercept']

        return point_x, point_y



from pathlib import Path
from urllib.request import urlopen, Request
from urllib.error import URLError

from geoformat.conf.error_messages import (
    path_not_valid,
    path_not_valid_file_exists_overwrite_is_false,
    path_http_not_valid,
    path_not_http,
    path_not_a_dir,
)


def add_extension_path(path, add_extension):
    """
    Add extension to path if add extension is specified

    :param path: input path
    :param add_extension:
    :return: output path
    """
    if add_extension and add_extension != path.suffix:
        path = path.with_suffix(path.suffix + add_extension)

    return path


def path_is_file(path):
    """
    Test if given path is a file.

    :param path: path to something
    :return: True if path is a path / False if not.
    """
    p = Path(path)
    is_file = False
    if p.is_file():
        is_file = True

    return is_file


def path_is_dir(path):
    """
    Test if given path is a dir.

    :param path: path to something
    :return: True if path is a dir path / False if not.
    """
    p = Path(path)
    is_dir = False
    if p.is_dir():
        is_dir = True

    return is_dir

def path_is_http(path, headers=None):
    """
    Test if given path is a http valid http link.

    :param path: path to something.
    :param headers: optionally you can add headers of http parameters
    :return: True if path is a http path / False if not.
    """
    if headers is None:
        headers = {}
    is_http = False
    resp_code = None
    if str(path).startswith("http"):
        req = Request(path, headers=headers)
        try:
            resp = urlopen(req)
            resp_code = resp.code
            if resp_code == 200:
                is_http = True
        except URLError:
            resp_code = 404

    return is_http, resp_code



def verify_input_path_is_file(path):
    """
    Transform str path to Path object from pathlib if given path exists and is a file
    :param path: str path or pathlib Path object
    :return: Path object or error message if path isn't valid.
    """

    if path_is_file(path) is True:
        p = Path(path)
        return p
    else:
        raise Exception(path_not_valid.format(path=path))

def verify_input_path_is_dir(path):
    """
    Transform str path to Path object from pathlib if given path exists and is a dir
    :param path: str path or pathlib Path object
    :return: Path object
    """
    p = Path(path)
    if p.is_dir():
        return p
    else:
        raise Exception(path_not_a_dir.format(path=path))

def verify_input_path_is_http(path, headers=None):
    """
    Take http path and test it validity. Return http path if it ok.

    :param path: str path or pathlib Path object
    :param headers: Optionally you can add headers of http parameters
    :return: http path if it's ok then error message
    """
    is_http, resp_code = path_is_http(path=path, headers=headers)
    if is_http is True:
        return path
    else:
        if resp_code is None:
            raise Exception(path_not_http.format(path=path))
        else:
            raise Exception(path_http_not_valid.format(path=path, code=resp_code))


def open_http_path(path, headers=None):
    """
    Open a http path and return response.

    :param path: http path
    :param headers: optionally you can add headers of http parameters.
    :return: response of http request
    """
    if headers is None:
        headers = {}
    req = Request(path, headers=headers)
    resp = urlopen(req)

    return resp

def path_to_file_path(path, geolayer_name, overwrite=True, add_extension=None):
    """
    Return verified file path for input path.
    Make difference between path dir and path file and return always

    :param path: dir path or file path to be checked.
    :param geolayer_name: name of geolayer (only used when path is dir path)
    :param overwrite: True we overwrite file if exists, False we return an error if file exists
    :param add_extension: if input path if file (without extension) or dir if add extension is True we add file
    extension specified ('geojson', 'shp', 'kml' ...)
    :return: output path
    """
    p = Path(path)
    file_path = None
    if p.exists() is True:
        # check if is dir
        if p.is_dir() is True:
            file_path = p.joinpath(geolayer_name)
        # chek if it's file
        elif p.is_file():
            file_path = p
    # file or dir does not exist
    else:
        if p.parent.exists():
            file_path = p
        else:
            raise Exception(path_not_valid.format(path=path))

    # add extension
    file_path = add_extension_path(path=file_path, add_extension=add_extension)

    # check overwrite
    if file_path.exists() and overwrite is False:
        raise Exception(
            path_not_valid_file_exists_overwrite_is_false.format(path=file_path)
        )

    return file_path



from geoformat.geoprocessing.measure.mesure_distance import (
    euclidean_distance,
    point_vs_segment_distance
)
from geoformat.geoprocessing.connectors.operations import segment_to_bbox

# predicate return :
# - a boolean for geometries confrontation
# - or an information about position between a geometry in relation to another

# Return BOOLEAN
# Point vs
# point_intersects_point
# point_intersects_segment
# point_intersects_bbox
#
# segment vs
# segment_intersects_bbox
# segment_intersects_segment
#
# bbox vs
# bbox_intersects_bbox

# Return POSITION
# point_position_segment
# ccw_or_cw_segments


def point_intersects_point(point_a, point_b, tolerance=None):
    """
    Return True if point_a and point_b have sames coordinates

    :param point_a:
    :param point_b:
    :param tolerance:
    :return:
    """
    if tolerance:
        intersects = False
        distance_point_a_to_point_b = euclidean_distance(point_a=point_a, point_b=point_b)
        if abs(distance_point_a_to_point_b - tolerance) <= tolerance:
            intersects = True
    else:
        intersects = point_a[0] == point_b[0] and point_a[1] == point_b[1]

    return intersects


# def point_intersects_segment(point, segment):
#     """
#     Test if point intersects a segment
#     """
#     if point_position_segment(point, segment) == 'ON':
#         return True
#     else:
#         return False


def point_intersects_segment(point, segment, tolerance=None):
    """
    Test if point intersects a segment
    Optionaly we add the consideration of tolerance to our calculations.
        Using tolerance is 3x more longer than without.

    How is used tolerance :
        - we transform point that we want to test to two segments that is perpendicular to input segment. These segments
         have a total length of twice the tolerance value.
        - we check if this new perpendiculars segment intersects input segment.

    :param point:
    :param segment:
    :param tolerance:
    :return:
    """

    if tolerance:
        intersects_value = False
        # because floating value we must define distance output rounding that match with tolerance rounding
        distance_point_to_segment = point_vs_segment_distance(point=point, segment=segment)
        if abs(distance_point_to_segment - tolerance) <= tolerance:
            intersects_value = True
    else:
        # test if point intersects segment basically
        if point_position_segment(point, segment) == 'ON':
            return True
        else:
            return False

    return intersects_value


def point_intersects_bbox(point, bbox):
    """
    This function send a boolean (true or false) that indicate if a given point intersect a given bbox

        Input:
            point
            bbox

        Output:
            True or False(boolean)
    """

    (point_x, point_y) = point
    (bbox_x_min, bbox_y_min, bbox_x_max, bbox_y_max) = bbox

    return bbox_x_min <= point_x <= bbox_x_max and bbox_y_min <= point_y <= bbox_y_max


def segment_intersects_segment(segment_a, segment_b):
    """"
    function that return True if two given segment intersects or False if not

    :param
    :segment_a: first segment to compare with second segment (segment_b)
    :segment_b: second segment to compare with first segment (segment_a)
    :return: True if intersects / False if not intersects
    """
    (point_a_1, point_a_2) = segment_a
    (point_b_1, point_b_2) = segment_b
    segment_a_1 = (point_a_2, point_b_1)
    segment_a_2 = (point_a_2, point_b_2)
    segment_b_1 = (point_b_2, point_a_1)
    segment_b_2 = (point_b_2, point_a_2)

    (pt_a_x, pt_a_y) = segment_a[0]
    (pt_b_x, pt_b_y) = segment_a[1]
    (pt_c_x, pt_c_y) = segment_a_1[1]

    a = (pt_b_x - pt_a_x) * (pt_c_y - pt_a_y)
    b = (pt_b_y - pt_a_y) * (pt_c_x - pt_a_x)
    if a > b:
        ccw_segment_a_1 = 'CCW'
    elif a < b:
        ccw_segment_a_1 = 'CW'
    else:
        ccw_segment_a_1 = False

    (pt_c_x, pt_c_y) = segment_a_2[1]

    a = (pt_b_x - pt_a_x) * (pt_c_y - pt_a_y)
    b = (pt_b_y - pt_a_y) * (pt_c_x - pt_a_x)
    if a > b:
        ccw_segment_a_2 = 'CCW'
    elif a < b:
        ccw_segment_a_2 = 'CW'
    else:
        ccw_segment_a_2 = False

    (pt_a_x, pt_a_y) = segment_b[0]
    (pt_b_x, pt_b_y) = segment_b[1]
    (pt_c_x, pt_c_y) = segment_b_1[1]

    a = (pt_b_x - pt_a_x) * (pt_c_y - pt_a_y)
    b = (pt_b_y - pt_a_y) * (pt_c_x - pt_a_x)
    if a > b:
        ccw_segment_b_1 = 'CCW'
    elif a < b:
        ccw_segment_b_1 = 'CW'
    else:
        ccw_segment_b_1 = False

    (pt_c_x, pt_c_y) = segment_b_2[1]

    a = (pt_b_x - pt_a_x) * (pt_c_y - pt_a_y)
    b = (pt_b_y - pt_a_y) * (pt_c_x - pt_a_x)
    if a > b:
        ccw_segment_b_2 = 'CCW'
    elif a < b:
        ccw_segment_b_2 = 'CW'
    else:
        ccw_segment_b_2 = False

    # If segments are parallels
    if ccw_segment_a_1 is False and ccw_segment_a_2 is False and ccw_segment_b_1 is False and ccw_segment_b_2 is False:
        segment_a_bbox = segment_to_bbox(segment_a)
        segment_b_bbox = segment_to_bbox(segment_b)
        return bbox_intersects_bbox(segment_a_bbox, segment_b_bbox)
    else:
        if ccw_segment_a_1 != ccw_segment_a_2:
            if ccw_segment_b_1 != ccw_segment_b_2:
                return True
        return False


def segment_intersects_bbox(segment, bbox):
    """
        Output :
            result (boolean) : True or False
    """
    point_start = segment[0]
    point_end = segment[1]
    (x_min, y_min, x_max, y_max) = bbox

    segment_bbox = segment_to_bbox(segment)
    # the bbox of segment and bbox must intersects
    if bbox_intersects_bbox(segment_bbox, bbox):
        # if begin or end point of segment is on bbox
        if point_intersects_bbox(point_start, bbox) or point_intersects_bbox(point_end, bbox):
            return True
        # if segment bbox intersect other bbox so segment intersect bbox :)
        else:
            segment_est = ((x_min, y_min), (x_min, y_max))
            segment_north = ((x_min, y_max), (x_max, y_max))
            segment_west = ((x_max, y_max), (x_max, y_min))
            segment_south = ((x_max, y_min), (x_min, y_min))
            # Each segment is test while result is true
            segments_list = (segment_est, segment_north, segment_west, segment_south)
            for segment_in_list in segments_list:
                if segment_intersects_segment(segment_in_list, segment):
                    return True
            return False
    else:
        return False


def bbox_intersects_bbox(bbox_a, bbox_b):
    """
    This function return a Truth Value Testing (True False) that qualifies the rectangle intersection


    True : bbox intersects
    False : bbox doesn't intersects

    The algorithm is inspired from : http://stackoverflow.com/questions/13390333/two-rectangles-intersection
                                    IA  : https://web.archive.org/web/*/http://stackoverflow.com/questions/13390333/two-rectangles-intersection

        Input :
            bbox_a : first boundary box
            bbox_b : second boundary box

        Output :
            result (boolean) : True or False

    """
    (x_min_a, y_min_a, x_max_a, y_max_a) = bbox_a
    (x_min_b, y_min_b, x_max_b, y_max_b) = bbox_b

    return x_min_a <= x_max_b and x_max_a >= x_min_b and y_min_a <= y_max_b and y_max_a >= y_min_b


def ccw_or_cw_segments(segment_a, segment_b):
    """
    Return orientation for two consecutive segments. The second coordinate in segment A must be the same
    that the first segment B coordinate.
    This orientation is given in relation to a north-south-east-west oriented axis.

    This function is inspired from  : https://www.toptal.com/python/computational-geometry-in-python-from-theory-to-implementation

    :param:
    :segment_a: (list/tuple) first segment (last point must be same that second segment (segment_b))
    :segment_b: (list/tuple) second segment (first point must be same that first segment (segment_a))
    :return: - 'CCW': counter clock wise
             - 'CW': clock wise
             - 'NEITHER' : same direction
    """
    if segment_a[1][0] == segment_b[0][0] and segment_a[1][1] == segment_b[0][1]:
        (pt_a_x, pt_a_y) = segment_a[0]
        (pt_b_x, pt_b_y) = segment_a[1]
        (pt_c_x, pt_c_y) = segment_b[1]

        slope_a = (pt_b_x - pt_a_x) * (pt_c_y - pt_a_y)
        slope_b = (pt_b_y - pt_a_y) * (pt_c_x - pt_a_x)

        if slope_a == slope_b:
            return 'NEITHER'  # |
        elif slope_a > slope_b:
            return 'CCW'  # ↺
        else:
            return 'CW'  # ↻


def point_position_segment(point, segment):
    """
    Return position of point regarding a segment
    :param:
    :point: point that we position
    :segment: segment that
    :return: - 'LEFT'
             - 'RIGHT'
             - 'ON'
             - 'NEITHER'
    """
    segment_b = (segment[1], point)
    orientation = ccw_or_cw_segments(segment, segment_b)
    if orientation in {"CW", 'CCW'}:
        if orientation == 'CCW':
            return 'LEFT'
        else:
            return 'RIGHT'
    else:
        seg_bbox = segment_to_bbox(segment)
        if point_intersects_bbox(point, seg_bbox):
            return 'ON'
        else:
            return 'NEITHER'


if __name__ == '__main__':

    from tests.geoformat.geoprocessing.connectors.test_predicates import point_intersects_segment_parameters
    import time

    iteration = 100000
    now = time.time()
    for i in range(iteration):
        for key in point_intersects_segment_parameters:
            point = point_intersects_segment_parameters[key]['point']
            segment = point_intersects_segment_parameters[key]['segment']
            point_vs_segment_distance(point, segment)

    print('distance seule', time.time() - now)

    now = time.time()
    for i in range(iteration):
        for key in point_intersects_segment_parameters:
            point = point_intersects_segment_parameters[key]['point']
            segment = point_intersects_segment_parameters[key]['segment']
            point_intersects_segment(point, segment)

    print("fonction sans tolerance 1", time.time() - now)

    now = time.time()
    # for i in range(iteration):
    #     for key in point_intersects_segment_parameters:
    #         point = point_intersects_segment_parameters[key]['point']
    #         segment = point_intersects_segment_parameters[key]['segment']
    #         tolerance = point_intersects_segment_parameters[key]['tolerance']
    #         point_intersects_segment_with_tolerance(point, segment, 0, tolerance_type='segment')

    print("fonction sans tolerance 2", time.time() - now)

    now = time.time()
    for i in range(iteration):
        for key in point_intersects_segment_parameters:
            point = point_intersects_segment_parameters[key]['point']
            segment = point_intersects_segment_parameters[key]['segment']
            tolerance = point_intersects_segment_parameters[key]['tolerance']
            point_intersects_segment(point, segment, 0)

    print("fonction sans tolerance 3", time.time() - now)


    now = time.time()
    # for i in range(iteration):
    #     for key in point_intersects_segment_parameters:
    #         point = point_intersects_segment_parameters[key]['point']
    #         segment = point_intersects_segment_parameters[key]['segment']
    #         tolerance = point_intersects_segment_parameters[key]['tolerance']
    #         point_intersects_segment_with_tolerance(point, segment, tolerance, tolerance_type='segment')

    print("tolerance segment", time.time() - now)

    now = time.time()
    for i in range(iteration):
        for key in point_intersects_segment_parameters:
            point = point_intersects_segment_parameters[key]['point']
            segment = point_intersects_segment_parameters[key]['segment']
            tolerance = point_intersects_segment_parameters[key]['tolerance']
            point_intersects_segment(point, segment, tolerance)

    print("tolerance distance", time.time() - now)



import copy

from geoformat.conf.error_messages import (
    field_missing,
    field_exists,
    field_type_not_valid,
    field_width_not_valid,
    field_precision_not_valid,
    variable_must_be_int,
    field_name_not_indexing,
    field_name_still_indexing,
)

from geoformat.conf.fields_variable import (
    field_metadata_width_required,
    field_metadata_precision_required,
    field_type_set,
)

from geoformat.conf.format_data import value_to_iterable_value


def drop_field_in_metadata(metadata, field_name_or_field_name_list):
    """
    Drop field(s) in metadata contains in field_name_or_field_name_list

    :param metadata:
    :param field_name_or_field_name_list:
    :return: metadata without deleted field
    """
    field_name_or_field_name_list = value_to_iterable_value(
        field_name_or_field_name_list, output_iterable_type=list
    )
    reorder_field_index = []
    if "fields" in metadata:
        output_metadata = copy.deepcopy(metadata)
        for field_name in field_name_or_field_name_list:
            if field_name in metadata["fields"]:
                if "index" in metadata["fields"][field_name]:
                    reorder_field_index.append(metadata["fields"][field_name]["index"])
                del output_metadata["fields"][field_name]

        # reorder field index
        if reorder_field_index:
            output_metadata["fields"] = reorder_metadata_field_index_after_field_drop(
                fields_metadata=output_metadata["fields"],
                reorder_field_index=reorder_field_index,
            )

        # if there is no field in metadata we delete fields key in it
        if len(output_metadata["fields"]) == 0:
            del output_metadata["fields"]
    else:
        output_metadata = metadata

    return output_metadata


def drop_field_that_not_exists_in_metadata(
    metadata,
    not_deleting_field_name_or_field_name_list
):
    """
    Drop field that not exists in metadata

    :param metadata: metadata with the fields you want to keep.
    :param not_deleting_field_name_or_field_name_list: field name (or field name list) that you want to keep
    :return: metadata without field
    """
    if metadata["fields"]:
        field_to_delete_list = []
        for field_name in metadata["fields"]:
            if field_name not in not_deleting_field_name_or_field_name_list:
                field_to_delete_list.append(field_name)

        if field_to_delete_list:
            metadata = drop_field_in_metadata(
                metadata=metadata, field_name_or_field_name_list=field_to_delete_list)

    return metadata


def reorder_metadata_field_index_after_field_drop(fields_metadata, reorder_field_index):
    """
    This function is used when a field is deleted in a geolayer.

    Reorder "index" key in fields_metadata for step(s) specified in reorder_field_index.

    :param fields_metadata: field metadata from geolayer
    :param reorder_field_index: list that contains original index to delete field.
    :return: fields_metadata with field index reordered after field delete
    """

    # init reorder field index
    reorder_field_index = value_to_iterable_value(reorder_field_index, list)

    # sort index
    reorder_field_index.sort()
    # loop on each index
    for idx in reversed(reorder_field_index):
        for field_name, field_metadata in fields_metadata.items():
            if field_metadata["index"] > idx:
                field_metadata["index"] = field_metadata["index"] - 1

    return fields_metadata


def rename_field_in_metadata(metadata, old_field_name, new_field_name):
    """
    Rename field in metadata

    :param old_field_name: current name of the field you want to modify.
    :param new_field_name: new name to the field that we want change.
    :return: metadata with rename field name.
    """
    output_metadata = copy.deepcopy(metadata)
    if geolayer_fields_metadata := output_metadata.get("fields", {}):
        if old_field_name in geolayer_fields_metadata:
            if new_field_name not in geolayer_fields_metadata:
                # rename in geolayer metadata
                geolayer_fields_metadata[new_field_name] = dict(
                    geolayer_fields_metadata[old_field_name]
                )
                del geolayer_fields_metadata[old_field_name]
            else:
                raise Exception(field_exists.format(field_name=new_field_name))
        else:
            raise Exception(field_missing.format(field_name=old_field_name))

    return output_metadata


def create_field_in_metadata(
    metadata, field_name, field_type, field_width, field_precision
):
    """
    Add a field in metadata

    :param field_name: new field name to create
    :param field_type: field type for new field
    :param field_width: field width for new field
    :param field_precision: field precision for new field
    :return: input metadata with new field in it
    """

    # check if field not exists
    if metadata_fields := metadata.get("fields"):
        if field_name in metadata_fields:
            raise Exception(field_exists.format(field_name=field_name))
    else:
        metadata["fields"] = {}

    # check field_type
    if field_type not in field_type_set:
        raise Exception(
            field_type_not_valid.format(
                field_type=field_type, field_type_list=sorted(list(field_type_set))
            )
        )

    # check field_width
    if field_type in field_metadata_width_required:
        if field_width is None:
            raise Exception(field_width_not_valid.format(field_name=field_name))
        else:
            if not isinstance(field_width, int):
                raise Exception(
                    variable_must_be_int.format(variable_name="field_width")
                )
    else:
        field_width = None

    # check field_precision
    if field_type in field_metadata_precision_required:
        if field_precision is None:
            raise Exception(field_precision_not_valid.format(field_name=field_name))
        else:
            if not isinstance(field_precision, int):
                raise Exception(
                    variable_must_be_int.format(variable_name="field_precision")
                )
    else:
        field_precision = None

    metadata["fields"][field_name] = {"type": field_type}

    if field_width:
        metadata["fields"][field_name]["width"] = field_width

    if field_precision:
        metadata["fields"][field_name]["precision"] = field_precision

    # add index
    metadata["fields"][field_name]["index"] = len(metadata["fields"]) - 1

    return metadata


def check_if_field_exists_in_metadata(metadata, field_name):
    """
    Check if field exists in given metadata.

    :param metadata: input metadata where existing field name is tested.
    :param field_name: name of field.
    :return: True if field exists False if not.
    """

    return_value = False
    if metadata.get("fields"):
        if metadata["fields"].get(field_name):
            return_value = True

    return return_value


def check_attributes_index_in_metadata(metadata, field_name, type=None):
    """
    Return True or False if an attributes index is found in metadata for input field_name optionaly we can test
    type index

    :param metadata: input metadata
    :param field_name: field name where is the index
    :param type: type of indexing (hashtable/btree ...)
    :return: Boolean
    """
    found_index = False
    if check_if_field_exists_in_metadata(metadata=metadata, field_name=field_name):
        if "index" in metadata:
            if "attributes" in metadata["index"]:
                if field_name in metadata["index"]["attributes"]:
                    found_index = True
                    if type:
                        if metadata["index"]["attributes"][field_name]['metadata']["type"] != type:
                            found_index = False
    else:
        raise Exception(field_missing.format(field_name=field_name))

    return found_index


def add_attributes_index_in_metadata(metadata, field_name, index):
    """Store index in metadata

    :param metadata: metadata to add an attributes
    :field_name: name of field concern by index
    :index: index that we want to add to field of geolayer
    :return: geolayer with index in it
    """

    # check if field_name exists
    if (
        check_if_field_exists_in_metadata(metadata=metadata, field_name=field_name)
        is False
    ):
        raise Exception(field_missing.format(field_name=field_name))

    # create index structure in metadata (if not)
    metadata_index = metadata.get("index")
    if metadata_index is None:
        metadata["index"] = {}
        metadata_index = metadata["index"]

    metadata_index_attributes = metadata_index.get("attributes")
    if metadata_index_attributes is None:
        metadata_index["attributes"] = {}
        metadata_index_attributes = metadata_index["attributes"]

    # check if index is yet in store in geolayer
    if metadata_index_attributes.get(field_name):
        raise Exception(field_name_still_indexing.format(field_name=field_name))

    metadata_index_attributes[field_name] = index

    return metadata


def delete_attributes_index_in_metadata(metadata, field_name, type):
    """Delete attributes index when existing in metadata. Opionnaly you can filter on index type.

    :param metadata: input metadata
    :param field_name: field name where is the index
    :param type: type of indexing (hashtable/btree ...)
    :return: Geolayer without index for given field_name (and optionally index type)
    """

    if check_attributes_index_in_metadata(metadata, field_name, type) is True:
        del metadata["index"]["attributes"][field_name]
        if len(metadata["index"]["attributes"]) == 0:
            del metadata["index"]["attributes"]
        if len(metadata["index"]) == 0:
            del metadata["index"]
    else:
        raise Exception(field_name_not_indexing.format(field_name=field_name))

    return metadata



from geoformat.conf.error_messages import field_width_float_not_none
from geoformat.manipulation.metadata_manipulation import (
    reorder_metadata_field_index_after_field_drop,
)
from geoformat.conf.fields_variable import python_type_to_geoformat_field_type


def geometries_scan_to_geometries_metadata(geometry_scan, crs=None, extent=True):
    """
    Create geometry metadata dict from geometry scan get from features_geometry_ref_scan.

    :param geometry_scan: result from features_geometry_ref_scan function
    :param crs: coordinates reference system of features contains in geolayer
    :param extent: if True add extent geolyer to geometry metadata
    :return: geometry metadata
    """
    if geometry_scan["type"]:
        geometry_metadata = {"type": geometry_scan["type"]}
        if extent and "extent" in geometry_scan:
            if geometry_scan["extent"]:
                geometry_metadata["extent"] = geometry_scan["extent"]
        if crs:
            geometry_metadata["crs"] = crs
    else:
        geometry_metadata = None

    return geometry_metadata


def from_field_scan_determine_field_width(
    field_type,
    field_in_list,
    field_width_str=None,
    field_width_float=None,
    field_width_list=None,
    max_width=None,
):
    """
    Determine the width of a field when you want to recast a field, with field scan information str/float/list.

    :param field_type: python field type wanted.
    :param field_in_list: True if field's value is in list False if not.
    :param field_width_str: result to field scan for field_width_str key.
    :param field_width_float: result to field scan for field_width_float key.
    :param field_width_list: result to field scan for field_width_list key.
    :param max_width: is it possible optionally to put a maximum value because in some format fields have a max width.
    :return: width for field metadata
    """
    if field_type == str:
        field_width = field_width_str
        # if we transform a list original to str (does not work with StringList type)
        if field_width_list is not None:
            if field_type is str and field_in_list is False:
                if field_width_list > field_width_str:
                    field_width = field_width_list

    # get with and precision for float
    elif field_type == float:
        if field_width_float is None:
            raise Exception(field_width_float_not_none)
        field_width = field_width_float
    else:
        field_width = None

    if field_width and max_width and field_width > max_width:
        field_width = max_width

    return field_width


def fields_scan_to_fields_metadata(fields_scan):
    """
    Deduce fields metadata from fields scan result.

    List of field's type and necessary keys :

        | type          | width    | precision | index    |
        |---------------|----------|-----------|----------|
        | 'Integer'     | None     | None      | Optional |
        | 'IntegerList' | None     | None      | Optional |
        | 'Real'        | Required | Required  | Optional |
        | 'RealList'    | Required | Required  | Optional |
        | 'String'      | Required | None      | Optional |
        | 'StringList'  | Required | None      | Optional |
        | 'Binary'      | None     | None      | Optional |
        | 'Date'        | None     | None      | Optional |
        | 'Time'        | None     | None      | Optional |
        | 'DateTime'    | None     | None      | Optional |
        | 'Boolean'     | None     | None      | Optional |

        example :
            field_metadata  = {
                'field_a': {'type': 'Integer', 'index': 0},
                'field_b': {'type': 'Real', 'width': 4, 'precision': 2, 'index': 1}
            }


    :param fields_scan: result of geolayer_fields_type_scan
    :return: geolayer fields metadata
    """
    # create fields metadata dict
    fields_metadata = {}

    # create field_name list ordered by 'field_index' in field_name data
    field_name_list = [None] * len(fields_scan)
    for i_field, (field_name, field_data) in enumerate(fields_scan.items()):
        # if "field_index" key in field_data we store field_name in field_name_list in 'field_index' order
        if "field_index" in field_data:
            field_idx = field_data["field_index"]
        else:
            field_idx = i_field
        field_name_list[field_idx] = field_name

    # loop on fields and create field metadata dict
    reorder_field_index = []
    for field_name in field_name_list:
        field_dict = fields_scan[field_name]
        if field_dict["field_delete"] is False:
            field_metadata_dict = {}
            # determine field type
            field_type = field_dict["field_type"]
            field_type_key = field_type
            if field_dict["field_list"] is True:
                field_type_key = (field_type_key, list)
            # add data in field_metadata_dict
            # get type
            field_metadata_dict["type"] = python_type_to_geoformat_field_type[
                field_type_key
            ]
            # get width for str
            if field_type == str:
                field_metadata_dict["width"] = field_dict["field_width_str"]
                # if we transform an list original to str (does not work with StringList type)
                if field_type is str and field_dict["field_list"] is False:
                    if (
                        field_dict["field_width_list"]
                        and field_dict["field_width_list"]
                        > field_dict["field_width_str"]
                    ):
                        field_metadata_dict["width"] = field_dict["field_width_list"]
            # get with and precision for float
            if field_type == float:
                field_metadata_dict["width"] = field_dict["field_width_float"] or 1
                field_metadata_dict["precision"] = field_dict["field_precision"]
            # get field_index (if require)
            if "field_index" in field_dict:
                field_metadata_dict["index"] = field_dict["field_index"]

            # write field metadata in fields metadata
            fields_metadata[field_name] = field_metadata_dict

        else:
            # if field must be deleted and fields have an index (we have to reorder it)
            if "field_index" in field_dict:
                reorder_field_index.append(field_dict["field_index"])

    # if we need to reorder field index
    if reorder_field_index:
        fields_metadata = reorder_metadata_field_index_after_field_drop(
            fields_metadata=fields_metadata, reorder_field_index=reorder_field_index
        )

    return fields_metadata


def get_field_name_list_ordered_by_i_field(fields_metadata):
    """
    Return field_name list ordered by i_field

    :param fields_metadata: field's metadata
    """
    if fields_metadata:
        field_name_list = [None] * len(fields_metadata)
        for i_field, field_name in enumerate(fields_metadata):
            if "index" in fields_metadata[field_name]:
                idx_field = fields_metadata[field_name]["index"]
            else:
                idx_field = i_field
            field_name_list[idx_field] = field_name
    else:
        field_name_list = []

    return field_name_list



import copy

from geoformat.conf.error_messages import metadata_fields_not_same, metadata_geometry_crs
from geoformat.conversion.feature_conversion import feature_serialize, feature_deserialize
from geoformat.conversion.metadata_conversion import get_field_name_list_ordered_by_i_field


def union_metadata(metadata_a, metadata_b, metadata_name, field_name_filter=None, feature_serialize=False):
    """

    """
    new_metadata = {"name": metadata_name}
    # check if field metadata are equal
    if 'fields' in metadata_a and 'fields' in metadata_b:

        if metadata_a['fields'].keys() != metadata_b['fields'].keys():
            raise Exception(metadata_fields_not_same)
        else:
            field_metadata_a = metadata_a['fields']
            field_metadata_b = metadata_b['fields']
            # get order field
            if field_name_filter:
                field_name_list = field_name_filter
            else:
                field_name_list = get_field_name_list_ordered_by_i_field(fields_metadata=field_metadata_a)
            for i_field, field_name in enumerate(field_name_list):
                new_metadata['fields'] = {}
                field_name_metadata_a = field_metadata_a[field_name]
                field_name_metadata_b = field_metadata_b[field_name]
                # type
                if field_type := field_name_metadata_a['type'] != field_name_metadata_b['type']:
                    raise Exception()
                else:
                    new_metadata['fields'][field_name] = {'type': field_type}
                # width
                if field_width_a := field_name_metadata_a.get('width'):
                    field_width_b = field_name_metadata_b['width']
                    field_width = max((field_width_a, field_width_b))
                    new_metadata['fields'][field_name]['width'] = field_width
                # precision
                if field_precision_a := field_name_metadata_a.get('precision'):
                    field_precision_b = field_name_metadata_b['precision']
                    field_precision = max((field_precision_a, field_precision_b))
                    new_metadata['fields'][field_name]['precision'] = field_precision
                # index
                new_metadata['fields'][field_name]['index'] = i_field
            new_metadata['fields'] = copy.deepcopy(metadata_a['fields'])

    # if geometry
    metadata_a_geometry_ref = metadata_a.get('geometry_ref')
    metadata_b_geometry_ref = metadata_b.get('geometry_ref')
    if metadata_a_geometry_ref is not None or metadata_b_geometry_ref is not None:
        metadata_a_geometry_ref_type = copy.deepcopy(metadata_a_geometry_ref.get('type', set()))
        metadata_b_geometry_ref_type = copy.deepcopy(metadata_b_geometry_ref.get('type', set()))
        new_metadata_geometry_type = metadata_a_geometry_ref_type.union(metadata_b_geometry_ref_type)

        new_metadata['geometry_ref'] = {"type": new_metadata_geometry_type}
        metadata_a_geometry_ref_crs = metadata_a_geometry_ref.get('crs')
        metadata_b_geometry_ref_crs = metadata_b_geometry_ref.get('crs')
        if metadata_a_geometry_ref_crs or metadata_b_geometry_ref_crs:
            if metadata_a_geometry_ref_crs and metadata_b_geometry_ref_crs is None:
                crs = metadata_a_geometry_ref_crs
            elif metadata_b_geometry_ref_crs is None and metadata_b_geometry_ref_crs:
                crs = metadata_b_geometry_ref_crs
            else:
                if metadata_a_geometry_ref_crs != metadata_b_geometry_ref_crs:
                    raise Exception(metadata_geometry_crs)
                else:
                    crs = metadata_b_geometry_ref_crs
            new_metadata['geometry_ref']['crs'] = crs

    # if feature serialize
    if feature_serialize is True:
        new_metadata['feature_serialize'] = feature_serialize

    return new_metadata


def union_geolayer(geolayer_list, geolayer_name, serialize=False):
    """
    Union geolayers with compatible fields structure and non mixed coordinates reference system
    """
    for i_geolayer, geolayer in enumerate(geolayer_list):
        geolayer_metadata = geolayer['metadata']
        if i_geolayer == 0:
            previ_metadata = geolayer_metadata
        else:
            merge_metadata = union_metadata(metadata_a=previ_metadata, metadata_b=geolayer_metadata,
                                            metadata_name=geolayer_name, feature_serialize=serialize)
            previ_metadata = geolayer_metadata

    output_geolayer = {"metadata": merge_metadata, "features": {}}

    output_i_feat = 0
    for geolayer in geolayer_list:
        for i_feat, feature in geolayer['features'].items():
            # serialize data
            if serialize is True and geolayer['metadata'].get('feature_serialize', False) is False:
                feature = feature_serialize(feature=feature)
            elif serialize is False and geolayer['metadata'].get('feature_serialize', False) is True:
                feature = feature_deserialize(serialized_feature=feature)

            output_geolayer["features"][output_i_feat] = feature
            output_i_feat += 1

    return output_geolayer



from geoformat.geoprocessing.measure.mesure_distance import (
    euclidean_distance_point_vs_segment
)


def shoelace_formula(ring_coordinates, absolute_value=True):
    """
    This function give the area of a boundary in the unit of coordinates.
    It is useful to use shoelace_formula to determine encoding sens of boundary coordinates
    (see ccw_or_cw_boundary)

    :param ring_coordinates: list of coordinates
    :param absolute_value: True / False
    :return: ring area (in coordinates unit)
    """
    nb_vertex = len(ring_coordinates)
    nb_edge = nb_vertex - 1

    l = [(ring_coordinates[i + 1][0] - ring_coordinates[i][0]) * (
            ring_coordinates[i + 1][1] + ring_coordinates[i][1]) for i in range(nb_edge)]

    if absolute_value:
        return abs(sum(l) / 2)
    else:
        return sum(l) / 2


def triangle_area(vertex_a, vertex_b, vertex_c):
    """
    Return the area of the triangle defined by points a, b, c.
    """
    ax, ay = vertex_a
    bx, by = vertex_b
    cx, cy = vertex_c
    return 0.5 * abs((bx - ax)*(cy - ay) - (by - ay)*(cx - ax))



import os.path
import sys

try:
    from osgeo import ogr, osr
    import_ogr_success = True
except ImportError:
    import_ogr_success = False

from geoformat.conf.error_messages import import_ogr_error

from geoformat.conf.driver_variable import (
    OGR_DRIVER_NAMES,
    OGR_FORMAT_ESRI_SHAPEFILE,
    OGR_FORMAT_KML,
    OGR_FORMAT_XLSX,
    OGR_FORMAT_CSV,
    OGR_FORMAT_GEOJSON,
    OGR_FORMAT_POSTGRESQL,
    OGR_FORMAT_MAPINFO,
    OGR_FORMAT_GML,
    field_driver_uncompatibility
)
from geoformat.conf.geometry_variable import (
    GEOMETRY_CODE_TO_GEOMETRY_TYPE,
    GEOMETRY_TYPE
)
from geoformat.conversion.geometry_conversion import (
    ogr_geometry_to_geometry,
    geometry_to_ogr_geometry
)
from geoformat.geoprocessing.connectors.predicates import bbox_intersects_bbox
from geoformat.geoprocessing.geoparameters.bbox import bbox_union
from geoformat.conf.fields_variable import (
    field_metadata_width_required,
    field_metadata_precision_required
)
from geoformat.conversion.feature_conversion import (
    feature_serialize,
    feature_deserialize
)


DATA_TYPE = {
    "OFSTBoolean": 1,
    "OFSTFloat32": 3,
    "OFSTInt16": 2,
    "OFSTNone": 0,
    "OFTBinary": 8,
    "OFTDate": 9,
    "OFTDateTime": 11,
    "OFTInteger": 0,
    "OFTInteger64": 12,
    "OFTInteger64List": 13,
    "OFTIntegerList": 1,
    "OFTReal": 2,
    "OFTRealList": 3,
    "OFTString": 4,
    "OFTStringList": 5,
    "OFTTime": 10,
    "OFTWideString": 6,
    "OFTWideStringList": 7
}

geoformat_field_type_to_ogr_field_type = {
    'Integer': 0,
    'IntegerList': 1,
    'Real': 2,
    'RealList': 3,
    'String': 4,
    'StringList': 5,
    'Binary': 8,
    'Date': 9,
    'Time': 10,
    'DateTime': 11,
    'Boolean': (0, 1)
}


ogr_field_type_to_geoformat_field_type = {
    8: 'Binary',
    9: 'Date',
    11: 'DateTime',
    0: 'Integer',
    12: 'Integer',
    13: 'IntegerList',
    1: 'IntegerList',
    2: 'Real',
    3: 'RealList',
    4: 'String',
    5: 'StringList',
    10: 'Time',
    6: 'String',
    7: 'StringList',
    (0, 1): 'Boolean',
    (0, 2): 'Integer',
    (2, 3): 'Real',
}


def ogr_layer_to_geolayer(
    path,
    layer_id_or_name=None,
    field_name_filter=None,
    driver_name=None,
    bbox_extent=True,
    bbox_filter=None,
    serialize=False,
    feature_limit=None,
    feature_offset=None
):
    """
    'ogr_layer_to_geolayer' permet d'ouvrier un fichier sig externe (attributaire ou géométrique), de le lire et de
    stocker des informations filtrées. Il faut lui renseigner un chemin vers le fichier (obligatoirement).

    :param path: chemin d'accès au fichier + nom fichier ; Format : string ; ex : "C:\foo\bar.shp"
    :param layer_id_or_name: identifiant ou nom table concernée ; Format : string ou integer ou none
    :param field_name_filter: liste noms des champs filtrant ; Format : StringList  ou none
    :param driver_name: format du fichier en lecture, le script peu le déterminer ; Format : string (majuscule) ou none
    :param bbox_extent: if you want to add bbox for each geometry and extent in metadata
    :param bbox_filter if wout want make a bbox filter when you import data
    :return: geolayer : layer au format geoformat basé sur le fichier path renseigné en entrée filtrées par field_name_filter
    """

    if import_ogr_success:
        # create output geolayer
        geolayer = {}

        # if driver name create it and create data source
        if driver_name:
            driver = ogr.GetDriverByName(driver_name)
            data_source = driver.Open(path)
        else:
            # or create datasource without driver
            data_source = ogr.Open(path)

        # open layer (table) - condition required for database
        if not layer_id_or_name:
            layer = data_source.GetLayer()
        else:
            if isinstance(layer_id_or_name, str):
                # if layer_id_or_name is string
                layer = data_source.GetLayerByName(layer_id_or_name)
            elif isinstance(layer_id_or_name, int):
                # # if layer_id_or_name is int
                layer = data_source.GetLayer(layer_id_or_name)
            else:
                raise Exception('layer_id_or_name must be a str or int value')

        # get layer definition
        layer_defn = layer.GetLayerDefn()

        # CREATE GEOLAYER METADATA
        # nom de la layer
        metadata = {'name': layer.GetName()}

        # creation du dictionnaire des métadonnées des champs et ajout du nombre de champ dans la layer
        fields_metadata = {}
        # boucle sur la structure des champs de la layer en entree
        if field_name_filter:
            # gestion de la correspondance de la case champs/field_name_filtres
            field_name_filter_up = [field_name.upper() for field_name in field_name_filter]
            # création liste vide pour stocker après le nom de champ initiaux de la table en entrée
            field_name_ori = []
            # récupération éléments pour chaque champ compris dans la layer definition

        if field_name_filter:
            # si field_name_filter on réinitialize l'ordre d'apparition des champs à 0
            i_field_filter = 0

        for i_field in range(layer_defn.GetFieldCount()):
            field_defn = layer_defn.GetFieldDefn(i_field)
            field_name = field_defn.GetName()
            field_type = field_defn.GetType()
            field_sub_type = field_defn.GetSubType()
            if field_sub_type:
                field_type = (field_type, field_sub_type)
            field_type = ogr_field_type_to_geoformat_field_type[field_type]
            field_width = field_defn.GetWidth()
            field_precision = field_defn.GetPrecision()

            # ecriture de la metadata des champs du filtre dans les metadonnees des champs
            write_metadata = True
            # vérification si field_name_filter renseigné
            if field_name_filter:
                # si field_name est dans field_name_filter (gestion de la casse)
                if field_name.upper() not in field_name_filter_up:
                    write_metadata = False
                else:
                    field_name_ori.append(field_name)

            if write_metadata:
                # create field metadata and add field type in it
                field_metadata = {
                    'type': field_type,
                }
                # add width (if required)
                if field_type in field_metadata_width_required:
                    field_metadata['width'] = field_width
                # add precision (if required)
                if field_type in field_metadata_precision_required:
                    field_metadata['precision']= field_precision
                # add index
                if field_name_filter:
                    field_index = i_field_filter
                    i_field_filter += 1
                else:
                    field_index = i_field
                field_metadata['index'] = field_index

                # add field to fields metadata dict
                fields_metadata[field_name] = field_metadata

        # write fields metadata
        field_name_list = None
        if fields_metadata:
            metadata['fields'] = fields_metadata
            if field_name_filter:
                field_name_list = field_name_ori
            else:
                field_name_list = list(metadata['fields'].keys())

        # if geometry
        if layer.GetGeomType() != 100:
            # add geometry metadata
            geometry_type = GEOMETRY_CODE_TO_GEOMETRY_TYPE[layer.GetGeomType()]
            metadata['geometry_ref'] = {'type': geometry_type}
            if layer.GetSpatialRef():
                metadata['geometry_ref']['crs'] = layer.GetSpatialRef().ExportToWkt()
            else:
                metadata['geometry_ref']['crs'] = None

            if bbox_extent:
                metadata['geometry_ref']['extent'] = None


        if serialize:
            metadata['feature_serialize'] = True

        # ajout des metadadonnees dans geolayer
        geolayer['metadata'] = metadata

        # creation des features indépendamment les unes des autres, création structure attributaire pour chaque feature
        geolayer['features'] = {}

        #
        # i_feat : feature_id in input layer
        # i_feat_writed : feature_id in output layer
        # write_feature : True|False say if a feature can or cannot be written this depending on
        # filters option setting :
        #                           bbox_filter
        #                           feature_limit
        #                           feature_offset
        i_feat_writed = 0
        # set who list unique geometry type in layer
        geom_type_set = set([])
        for i_feat, feature in enumerate(layer):
            # init
            write_feature = True
            # test feature offset and limit
            if feature_offset:
                if i_feat < feature_offset:
                    write_feature = False

            if feature_limit:
                # stop loop if feature_limit is reached
                if i_feat_writed == feature_limit:
                    break

            # create empty feature
            new_feature = {}
            if write_feature:

                ################################################################################################################
                #
                #   geometry
                #
                ################################################################################################################

                # test if geom in feature
                geom = feature.GetGeometryRef()
                if geom:
                    # get geometry type
                    if geom.GetGeometryType() != geoformat_geom_type_to_ogr_geom_type(
                            metadata['geometry_ref']['type']) or GEOMETRY_CODE_TO_GEOMETRY_TYPE[geom.GetGeometryType()] \
                            not in geom_type_set:
                        geom_type_set = set(
                            [GEOMETRY_CODE_TO_GEOMETRY_TYPE[geom.GetGeometryType()]] + list(geom_type_set))

                    # recuperation des geometries / ajout dans le dictionnaire des features
                    temp_bbox_extent = False
                    if bbox_filter is not None:
                        temp_bbox_extent = True

                    # bbox and extent must be computed
                    if bbox_extent or temp_bbox_extent:
                        geometry = ogr_geometry_to_geometry(geom, True)
                        if bbox_extent:
                            # modify extent in metadata
                            if geolayer['metadata']['geometry_ref']['extent'] is None:
                                geolayer['metadata']['geometry_ref']['extent'] = geometry['bbox']
                            else:
                                extent_bbox = geolayer['metadata']['geometry_ref']['extent']
                                extent_bbox = bbox_union(extent_bbox, geometry['bbox'])
                                geolayer['metadata']['geometry_ref']['extent'] = extent_bbox
                    else:
                        geometry = ogr_geometry_to_geometry(geom, False)

                    # check if geometry bbox intersects bbox filter
                    if bbox_filter:
                        feat_bbox = geometry['bbox']
                        if bbox_intersects_bbox(bbox_filter, feat_bbox):
                            if bbox_extent:
                                new_feature['geometry'] = geometry
                            else:
                                del geometry['bbox']
                                new_feature['geometry'] = geometry
                        else:
                            write_feature = False
                    else:
                        new_feature['geometry'] = geometry

            ################################################################################################################
            #
            #   attributes
            #
            ################################################################################################################
            if write_feature:
                if field_name_list:
                    # ajout des données attributaires
                    new_feature['attributes'] = {}
                    # recuperation des informations attributaires pour les features
                    # si option filtre sur champ
                    for field_name in field_name_list:
                        if geolayer['metadata']['fields'][field_name]['type'] == "Binary":
                            field_value = feature.GetFieldAsBinary(field_name)
                        else:
                            field_value = feature.GetField(field_name)
                        if geolayer['metadata']['fields'][field_name]['type'] == "Boolean":
                            field_value = bool(field_value)
                        new_feature['attributes'][field_name] = field_value

                if serialize:
                    new_feature = feature_serialize(new_feature)

                geolayer['features'][i_feat_writed] = new_feature
                i_feat_writed += 1

        ## Check layer metadata

        # GEOMETRY METADATA
        # if there is a difference between layer metadata geom type and scan
        if 'geometry_ref' in metadata:
            if geom_type_set != set(metadata['geometry_ref']['type']):
                if len(geom_type_set) == 0:
                    # There is no goemetry then we delete geometry_ref metadata
                    del metadata['geometry_ref']
                elif len(geom_type_set) == 1:
                    metadata['geometry_ref']['type'] = list(geom_type_set)[0]
                else:
                    metadata['geometry_ref']['type'] = list(geom_type_set)

        return geolayer

    else:
        raise Exception(import_ogr_error)


def ogr_layers_to_geocontainer(
        path,
        field_name_filter=None,
        driver_name=None,
        bbox_extent=True,
        bbox_filter=None,
        feature_limit=None,
        serialize=False
):
    """
    'ogr_layers_to_geocontainer' crée une géodatasource comprenant des géolayers (format geoformat). Elle requiere
    la fonction layer_to_geoformat car elle boucle sur la fonction layer_to_geoformat, ce qui permet de récupérer les
    différentes géolayer et de les encapsuler dans une datasource.

    :param path: chemin d'accès au fichier + nom fichier ; Format : string ou list
    :param field_name_filter: liste noms des champs filtrant, même filtre pour tous; Format : StringList  ou none
    :param driver_name: format(s) fichier en lecture, script peu le déterminer ; Format : string (majuscule) ou liste ou none
    :param bbox_extent: if you want to add bbox for each geometry and extent in metadata
    :param bbox_filter: if you want filter input feature with a given bbox


    :return: géodatasource : un conteneur de layer au geoformat, filtrées par le fiel_name_filter
    """

    # fonction qui permet de boucler sur la 'layer_to_geoformat'
    def loop_list_layer(
        path,
        field_name_filter=None,
        driver_name=None,
        bbox_extent=None,
        bbox_filter=None,
        serialize=False
    ):
        """
        'loop_list_layer' permet de lancer en boucle la fonction ogr_layer_to_geolayer.

        :param path: chemin d'un fichier
        :param field_name_filter: liste noms des champs filtrant
        :param driver_name: format fichier en lecture
        :return: yield de géolayer
        """

        if driver_name:
            # si driver_name renseigné
            driver = ogr.GetDriverByName(driver_name)
            data_source = driver.Open(path)


        else:
            # détection interne du driver via la fonction ogr.open()
            data_source = ogr.Open(path)

        # lancement de la fonction ogr_layer_to_geolayer() et récupération des layers au fur et à mesure
        for layer_id, layer in enumerate(data_source):
            geolayer = ogr_layer_to_geolayer(path, layer_id, field_name_filter=field_name_filter,
                                             driver_name=driver_name, bbox_extent=bbox_extent, bbox_filter=bbox_filter,
                                             feature_limit=feature_limit, serialize=serialize)
            yield geolayer

    # création du conteneur de layers
    geocontainer = {'layers': {}, 'metadata': {}}

    # init parameters
    temp_layer_path, temp_field_name_filter, temp_driver_name, temp_bbox_extent, temp_bbox_filter = None, None, None, None, None

    # test si le path est une liste, si oui : boucle pour chaque élément de la liste
    if isinstance(path, str):
        path = [path]

    for i_path, temp_layer_path in enumerate(path):

        if not isinstance(temp_layer_path, str):
            sys.exit('path must be a string')

        temp_field_name_filter = None
        # test si field_name_filter renseigné
        if field_name_filter:
            if isinstance(field_name_filter[i_path], str):
                temp_field_name_filter = field_name_filter
            else:
                temp_field_name_filter = field_name_filter[i_path]
        temp_driver_name = None

        # test driver_name
        if driver_name:
            temp_driver_name = driver_name[i_path]
            if isinstance(driver_name, list):
                temp_driver_name = driver_name[i_path]
            else:
                temp_driver_name = driver_name

        # test bbox extent
        if bbox_extent:
            if isinstance(bbox_extent, list):
                temp_bbox_extent = bbox_extent[i_path]
            else:
                temp_bbox_extent = bbox_extent

        # si bbox filter
        if bbox_filter:
            if isinstance(bbox_extent, list):
                temp_bbox_filter = bbox_extent[i_path]
            else:
                temp_bbox_filter = bbox_filter

        # lancement de la fonction loop_list_layer
        for i_geolayer, geolayer in enumerate(
                loop_list_layer(temp_layer_path, temp_field_name_filter, temp_driver_name, temp_bbox_extent,
                                temp_bbox_filter, serialize)):
            # stockage des returns yield de la fonction loop
            geolayer_name = geolayer['metadata']['name']
            geocontainer['layers'][geolayer_name] = geolayer
            if temp_bbox_extent:
                geolayer_extent = geolayer['metadata']['geometry_ref']['extent']
                if i_geolayer == 0:
                    geocontainer_extent = geolayer_extent
                else:
                    geocontainer_extent = bbox_union(geocontainer_extent, geolayer_extent)
                geocontainer['metadata']['extent'] = geocontainer_extent

    return geocontainer


def geolayer_to_ogr_layer(
        geolayer,
        path,
        driver_name,
        ogr_options=None,
        order_fields=True,
):
    """
    This function allow to convert a geolayer to an other GIS format with GDAL/OGR library.

    Supported format :
        - esri shapefile
        - geojson
        - kml
        - postgresql

    WARNING : this function does not modify geolayer to be compatible with output format. You have to do that by
    yourself. However, this function verbally lists incompatibilities (wrong field type, incompatible geometry type
    combination) for the desired output format.

    :param geolayer: input geolayer that we want to convert.
    :param path: export output path (can be a file / a directory or a link to database).
    :param driver_name: output format.
    :param ogr_options: OGR/GDAL allow some options that you can use, example for postgres ogr_options=['OVERWRITE=YES']
    :param order_fields: True
    """

    if import_ogr_success:
        # création de l'ensemble des informations pour créer un fichier au format SIG
        # création d'un driver
        driver_name = driver_name.upper()
        if driver_name not in OGR_DRIVER_NAMES:
            raise ValueError("The given driver name {} is not correct.".format(driver_name))
        driver = ogr.GetDriverByName(driver_name)

        # récupération du path en 2 parties : la racine et l'extension
        (root, ext) = os.path.splitext(path)

        # test si il y a pas une extension
        if ext == '':
            # alors c'est un dossier ou l'adresse d'une base de données
            # recupération du nom de la layer
            layer_name = geolayer['metadata']['name']
            # si le chemin est bien un dossier existant
            if os.path.isdir(root):
                # récupération de l'extension suivant le driver_name
                # TODO add other drivers
                if driver_name == OGR_FORMAT_ESRI_SHAPEFILE:
                    # if geometry in geolayer
                    if 'geometry_ref' in geolayer["metadata"]:
                        new_ext = '.shp'
                    # else we write only dbf
                    else:
                        new_ext = '.dbf'
                elif driver_name == OGR_FORMAT_KML:
                    new_ext = '.kml'
                elif driver_name == OGR_FORMAT_XLSX:
                    new_ext = '.xlsx'
                elif driver_name== OGR_FORMAT_CSV:
                    new_ext = '.csv'
                elif driver_name == OGR_FORMAT_GEOJSON:
                    new_ext = '.geojson'
                else:
                    sys.exit('format non pris en compte')
                new_path = os.path.join(root, layer_name + new_ext)
            else:
                # Then we suppose that is a datasource
                if ogr.Open(root) is not None:
                    if driver_name == OGR_FORMAT_POSTGRESQL:
                        new_path = path
                    else:
                        raise ValueError("Given file is a datasource but the driver_name is not {}".format(
                            OGR_FORMAT_POSTGRESQL))
                else:
                    raise FileNotFoundError("Your path does not exists or is invalid")
            data_source = driver.CreateDataSource(new_path)
        else:
            if driver_name == OGR_FORMAT_XLSX:
                # on test s'il existe
                data_source = driver.Open(path, 1)
                if not data_source:
                    data_source = driver.CreateDataSource(path)
            else:
                # alors c'est un fichier
                data_source = driver.CreateDataSource(path)

        # récupération dans geolayer des informations nécessaires à la création d'une layer : nom, projection, geometry_type
        layer_name = geolayer['metadata']['name']

        # init
        # coordinates reference system (crs) information
        layer_crs = None
        # layer geometry type
        layer_ogr_geom_type = geoformat_geom_type_to_ogr_geom_type(geoformat_geom_type=None)

        if 'geometry_ref' in geolayer['metadata']:
            layer_crs = None
            if 'crs' in geolayer['metadata']['geometry_ref']:
                crs_data = geolayer['metadata']['geometry_ref']['crs']
                # from EPSG
                if crs_data is None:
                    layer_crs = None
                elif isinstance(crs_data, int):
                    layer_crs = osr.SpatialReference()
                    layer_crs.ImportFromEPSG(crs_data)
                # from OGC WKT
                elif isinstance(crs_data, str):
                    try:
                        layer_crs = osr.SpatialReference(crs_data)
                    except:
                        # si le format n'est pas reconnu tant pis pas de ref spatiale
                        print('WARNING: projection not recognized')
                else:
                    print('WARNING: crs value must be an ESPG code or a OGC WKT projection')

            if 'type' in geolayer['metadata']['geometry_ref']:
                layer_ogr_geom_type = verify_geom_compatibility(driver_name, geolayer['metadata']['geometry_ref']['type'])
            else:
                raise ValueError("A type shall be found in geometry_ref.")

        if 'feature_serialize' in geolayer['metadata']:
            serialize = geolayer['metadata']['feature_serialize']
        else:
            serialize = False

        # Create Layer
        if ogr_options:
            layer = data_source.CreateLayer(layer_name,
                                            srs=layer_crs,
                                            geom_type=layer_ogr_geom_type,
                                            options=ogr_options)
        else:
            layer = data_source.CreateLayer(layer_name,
                                            srs=layer_crs,
                                            geom_type=layer_ogr_geom_type)


        # création des fields (structure du fichier)
        # si l'on souhaite que l'ordre d'apparition des champs soit conservée
        if 'fields' in geolayer['metadata']:
            if order_fields:
                field_name_list = [None] * len(geolayer['metadata']['fields'])
                for i_field, field_name in enumerate(geolayer['metadata']['fields']):
                    field_name_list[geolayer['metadata']['fields'][field_name].get('index', i_field)] = field_name
            else:
                field_name_list = list(geolayer['metadata']['fields'].keys())

            # check if field type is compatible with output format
            field_type_error = []
            for field_name in field_name_list:
                # récupération des informations nécessaire à la création des champs
                geoformat_field_type = geolayer['metadata']['fields'][field_name]['type']
                driver_name_field_type_combination = (driver_name, geoformat_field_type)
                if driver_name_field_type_combination in field_driver_uncompatibility:
                    field_type_error.append(field_driver_uncompatibility[driver_name_field_type_combination].format(
                        field_name=field_name,
                        ogr_format=driver_name)
                    )

            if field_type_error:
                error_message = '\n'
                for error in field_type_error:
                    error_message += '\t' + error +'\n'
                raise Exception(error_message)

            ogr_layer_field_type = {}
            for field_name in field_name_list:
                # get necessary data to create field
                geoformat_field_type = geolayer['metadata']['fields'][field_name]['type']
                field_type = geoformat_field_type_to_ogr_field_type[geoformat_field_type]
                ogr_layer_field_type[field_name] = field_type
                if isinstance(field_type, tuple):
                    (field_type, field_sub_type) = field_type
                else:
                    field_sub_type = 0
                field_width = geolayer['metadata']['fields'][field_name].get('width', 0)
                field_precision = geolayer['metadata']['fields'][field_name].get('precision', 0)

                # création de la définition du champ (type, longueur, precision
                field = ogr.FieldDefn(field_name, field_type)
                field.SetSubType(field_sub_type)
                # ogr does not allow precision at 0 for real and realList
                if field_type in {2, 3} and field_precision == 0:
                    field_width += 1
                    field_precision += 1

                field.SetWidth(field_width)
                field.SetPrecision(field_precision)

                # création du champ
                layer.CreateField(field)

            # creation table de correspondance [au cas où la taille des champs est réduite lors de la création de la layer]
            # example DBF = 10 char maximum
            # if layerDefn() is define
            try:
                ct_field_name = {}
                for i in range(layer.GetLayerDefn().GetFieldCount()):
                    ct_field_name[field_name_list[i]] = layer.GetLayerDefn().GetFieldDefn(i).GetName()
            except:
                ct_field_name = {field_name: field_name for field_name in field_name_list}

        # création des features
        for i_feat in geolayer['features']:
            try:
                feature_ogr = ogr.Feature(layer.GetLayerDefn())
            except:
                feature_ogr = ogr.Feature()

            feature_geoformat = geolayer['features'][i_feat]

            if serialize:
                feature_geoformat = feature_deserialize(feature_geoformat)

            # if geometry in feature
            if 'geometry' in feature_geoformat:
                # get geometry
                geometry = feature_geoformat['geometry']
                # transform geometry to ogr geometry
                geom_ogr = geometry_to_ogr_geometry(geometry)
                # add geometry in ogr feature object
                feature_ogr.SetGeometry(geom_ogr)

            # test if attributes data in feature
            if 'attributes' in feature_geoformat:
                # loop on each layer [metadata] fields
                for field_name in geolayer['metadata']['fields']:
                    # we recuperate true field name [in case it has been truncated]
                    true_field_name = ct_field_name[field_name]
                    value_field = feature_geoformat['attributes'].get(field_name)
                    # if field value exists and is not NULL
                    if value_field is not None:
                        # convert binary field to hex and write it with SetFieldBinaryFromHexString
                        if ogr_layer_field_type[field_name] == 8:
                            value_field = value_field.hex()
                            feature_ogr.SetFieldBinaryFromHexString(true_field_name, value_field)
                        else:
                            # write data if error change field_value to string
                            try:
                                feature_ogr.SetField(true_field_name, value_field)
                            except (NotImplementedError, TypeError):
                                feature_ogr.SetField(true_field_name, str(value_field))

            layer.CreateFeature(feature_ogr)

        data_source.Destroy()
    else:
        raise Exception(import_ogr_error)


def geocontainer_to_ogr_data_source(
        geocontainer,
        path,
        driver_name,
        export_layer_list=None,
        ogr_options=None,
        order_fields=False
    ):
    """
    'geocontainer_to_ogr_data_source' est une procedure qui permet d'exporter une sélection ou l'ensemble des layers d'une
    datasource aux formats voulus. Le path renseignée peut être un dossier, une datasource, ou un fichier. On peut
    renseigner une liste ou un nom de layer pour filtrer l'export.

    :param geocontainer: la géodatasource complete
    :param path: chemin où aller sauvegarder (il peut être une liste ou un str)
    :param driver_name: le nom du drive peut etre une liste ou un seul qu'on applique à tous
    :param export_layer_list: liste des layers de la datasource a exporter, peut être list ou str. Si export_layer_list
    non rempli alors on exporte toutes les layers, si export layer_list rempli = on exporte que ces layers là.
    Variable possible pour cette list : 'tc', 'ref_a', 'ref_b' (seulement)
    """

    if import_ogr_success:
        if export_layer_list:
            # test si il y a une liste des layers à exporter
            if isinstance(export_layer_list, list):
                for i_layer, export_layer_name in enumerate(export_layer_list):
                    # test si la layer fait partie de la liste des layers à sauvegarder
                    if export_layer_name in geocontainer['layers'].keys():
                        geolayer = geocontainer['layers'][export_layer_name]
                        if isinstance(path, list):
                            # path = fichier en dur
                            if isinstance(driver_name, list):
                                geolayer_to_ogr_layer(geolayer, path[i_layer], driver_name[i_layer],
                                                      ogr_options=ogr_options[i_layer], order_fields=order_fields)
                            else:
                                geolayer_to_ogr_layer(geolayer, path[i_layer], driver_name, ogr_options=ogr_options,
                                                      order_fields=order_fields)
                        else:
                            # path = dossier ou database
                            if isinstance(driver_name, list):
                                geolayer_to_ogr_layer(geolayer, path, driver_name[i_layer],
                                                      ogr_options=ogr_options[i_layer], order_fields=order_fields)
                            else:
                                geolayer_to_ogr_layer(geolayer, path, driver_name, ogr_options=ogr_options,
                                                      order_fields=order_fields)

            # si export_layer_list n'est pas une liste, elle contient qu'une valeur
            elif isinstance(export_layer_list, str):
                geolayer = geocontainer['layers'][export_layer_list]
                geolayer_to_ogr_layer(geolayer, path, driver_name, ogr_options=ogr_options, order_fields=order_fields)


        # si export_layer_list=None alors on exporte l'ensemble des layers de la geocontainer
        else:
            for i_layer, layer_name in enumerate(geocontainer['layers']):
                geolayer = geocontainer['layers'][layer_name]
                # si le path est une liste :
                if isinstance(path, list):
                    # et si le driver_name est une liste
                    if isinstance(driver_name, list):
                        geolayer_to_ogr_layer(geolayer, path[i_layer], driver_name[i_layer],
                                              ogr_options=ogr_options[i_layer], order_fields=order_fields)
                    # si non utiliser toujours le même driver
                    else:
                        geolayer_to_ogr_layer(geolayer, path[i_layer], driver_name, ogr_options=ogr_options,
                                              order_fields=order_fields)
                # sinon utiliser le même dossier
                else:
                    # test sur le driver_name pour voir lequel on donne
                    if isinstance(driver_name, list):
                        geolayer_to_ogr_layer(geolayer, path, driver_name[i_layer], ogr_options=ogr_options[i_layer],
                                              order_fields=order_fields)
                    else:
                        geolayer_to_ogr_layer(
                            geolayer,
                            path,
                            driver_name,
                            ogr_options=ogr_options,
                            order_fields=order_fields
                        )
    else:
        raise Exception(import_ogr_error)


def geoformat_geom_type_to_ogr_geom_type(geoformat_geom_type):
    """
    Make transformation between Geoformat geometry type to OGR geometry type

    :param geoformat_geom_type: (str) Geoformat geometry type
    :return: (int) ogr geom type correspondance
    """
    if geoformat_geom_type is None:
        return GEOMETRY_TYPE["None"]
    if geoformat_geom_type.upper() == 'GEOMETRY':  # Geometry
        return GEOMETRY_TYPE['Unknown']
    if geoformat_geom_type.upper() == 'POINT25D':  # Point25D
        return GEOMETRY_TYPE["Point25D"]
    if geoformat_geom_type.upper() == 'LINESTRING25D':  # LineString25D
        return GEOMETRY_TYPE["LineString25D"]
    if geoformat_geom_type.upper() == 'POLYGON25D':  # Polygon25D
        return GEOMETRY_TYPE["Polygon25D"]
    if geoformat_geom_type.upper() == 'MULTIPOINT25D':  # MultiPoint25D
        return GEOMETRY_TYPE["MultiPoint25D"]
    if geoformat_geom_type.upper() == 'MULTILINESTRING25D':  # MultiLineString25D
        return GEOMETRY_TYPE["MultiLineString25D"]
    if geoformat_geom_type.upper() == 'MULTIPOLYGON25D':  # MultiPolygon25D
        return GEOMETRY_TYPE["MultiPolygon25D"]
    if geoformat_geom_type.upper() == 'POINT':  # Point
        return GEOMETRY_TYPE["Point"]
    if geoformat_geom_type.upper() == 'LINESTRING':  # LineString
        return GEOMETRY_TYPE["LineString"]
    if geoformat_geom_type.upper() == 'POLYGON':  # Polygon
        return GEOMETRY_TYPE["Polygon"]
    if geoformat_geom_type.upper() == 'MULTIPOINT':  # MultiPoint
        return GEOMETRY_TYPE["MultiPoint"]
    if geoformat_geom_type.upper() == 'MULTILINESTRING':  # MultiLineString
        return GEOMETRY_TYPE["MultiLineString"]
    if geoformat_geom_type.upper() == 'MULTIPOLYGON':  # MultiPolygon
        return GEOMETRY_TYPE["MultiPolygon"]
    if geoformat_geom_type.upper() == 'GEOMETRYCOLLECTION':  # GeometryCollection
        return GEOMETRY_TYPE["GeometryCollection"]
    if geoformat_geom_type.upper() == 'NO GEOMETRY':  # No Geometry
        return GEOMETRY_TYPE["None"]


def verify_geom_compatibility(driver_name, metadata_geometry_type):
    """

    OGR Geometry Type List :
        -2147483647: 'Point25D'
        -2147483646: 'LineString25D'
        -2147483645: 'Polygon25D'
        -2147483644: 'MultiPoint25D'
        -2147483643: 'MultiLineString25D'
        -2147483642: 'MultiPolygon25D'
                  0: 'Geometry'
                  1: 'Point'
                  2: 'LineString'
                  3: 'Polygon'
                  4: 'MultiPoint'
                  5: 'MultiLineString'
                  6: 'MultiPolygon'
                  7: 'GeometryCollection'
                100: 'No Geometry'

    [[0],[1], [2, 5], [3, 6], [4], [100]],  # 'Esri Shapefile'
    [[1, 4], [2, 5], [3, 6], [100]],  # TAB 'Mapinfo File'
    [[1, 4], [2, 5], [3, 6], [100]],  # MIF/MID 'Mapinfo File'
    [[1, 2, 3, 4, 5, 6, 7, 100]],  # KML
    [[1, 2, 3, 4, 5, 6, 7, 100]],  # GML
    [[0, 1, 2, 3, 4, 5, 6, 7, 100]],  # GeoJSON
    [[1], [2], [3, 6], [4], [5], [100]],  # Geoconcept
    [[1], [2, 5], [3, 6], [4], [100]],  # FileGDB
    [[1, 2, 3, 4, 5, 6, 7, 100]],  # SQLite
    [[1, 2, 3, 4, 5, 6, 7, 100]],  # POSTGRESQL
    [[1, 2, 3, 4, 5, 6, 7, 100]]   # CSV
    ],

    """

    if isinstance(metadata_geometry_type, str):
        metadata_geometry_type = [metadata_geometry_type.upper()]
    elif isinstance(metadata_geometry_type, (tuple, list, set)):
        metadata_geometry_type = [value.upper() for value in metadata_geometry_type]

    set_geometry_type = set(metadata_geometry_type)

    if driver_name.upper() == OGR_FORMAT_ESRI_SHAPEFILE:
        # POLYGON
        set_polygon = {'NO GEOMETRY', 'POLYGON', 'MULTIPOLYGON'}
        if len(set_geometry_type) <= 3 and len(set_geometry_type.difference(set_polygon)) == 0:
            return 3
        # LINESTRING
        set_linestring = {'NO GEOMETRY', 'LINESTRING', 'MULTILINESTRING'}
        if len(set_geometry_type) <= 3 and len(set_geometry_type.difference(set_linestring)) == 0:
            return 2
        # POINT
        set_point = {'NO GEOMETRY', 'POINT', 'MULTIPOINT'}
        if len(set_geometry_type) <= 3 and len(set_geometry_type.difference(set_point)) == 0:
            return 1

    if driver_name.upper() == OGR_FORMAT_POSTGRESQL:
        if len(set_geometry_type) > 1:
            return 0
        else:
            return geoformat_geom_type_to_ogr_geom_type(metadata_geometry_type[0])

    if driver_name.upper() == OGR_FORMAT_GEOJSON:
        return 0

    if driver_name.upper() == OGR_FORMAT_CSV:
        return 100

    if driver_name.upper() == OGR_FORMAT_GML:
        if len(set_geometry_type) == 1:
            return geoformat_geom_type_to_ogr_geom_type(geoformat_geom_type=metadata_geometry_type[0])
        else:
            raise Exception('mixed geometry type in geolayer. driver {driver_name} only accept on type of geometry.'
                            ' You can use multi_geometry_to_single_geometry_geolayer to only work with single geometry '
                            'use'.format(driver_name=OGR_FORMAT_GML))

    if driver_name.upper() == OGR_FORMAT_MAPINFO:

        # POLYGON
        set_polygon = {'POLYGON', 'MULTIPOLYGON'}
        if len(set_geometry_type) <= 3 and len(set_geometry_type.difference(set_polygon)) == 0:
            return 3
        # LINESTRING
        set_linestring = {'LINESTRING', 'MULTILINESTRING'}
        if len(set_geometry_type) <= 3 and len(set_geometry_type.difference(set_linestring)) == 0:
            return 2
        # POINT
        set_point = {'POINT', 'MULTIPOINT'}
        if len(set_geometry_type) <= 3 and len(set_geometry_type.difference(set_point)) == 0:
            return 1

        # NO GEOMETRY
        if len(set_geometry_type) == 1 and 'NO GEOMETRY' in set_geometry_type:
            return 100




def envelope_to_bbox(envelope):
    """
    Convert envelope to bbox
        FYI :
            envelope format (x_min, x_max, y_min, y_max)
            bbox format (x_min, y_min, x_max, y_max)

    :param envelope: input envelope
    :return: (x_min, y_min, x_max, y_max)
    """

    return envelope[0], envelope[2], envelope[1], envelope[3]


def bbox_to_envelope(bbox):
    """
    Convert bbox to envelope
        FYI :
            envelope format (x_min, x_max, y_min, y_max)
            bbox format (x_min, y_min, x_max, y_max)

    :param bbox: input bbox
    :return: (x_min, x_max, y_min, y_max)
    """

    return bbox[0], bbox[2], bbox[1], bbox[3]


def bbox_extent_to_2d_bbox_extent(bbox_extent):
    """
    Convert a bbox or extent with 2 or more dimension to 2d extent

    :param bbox_extent:  bbox or extent with 2d or more dimension
    :return:  new bbox or extent with only 2 dimensions
    """
    mid_idx = int(len(bbox_extent) / 2)
    bbox = (bbox_extent[0], bbox_extent[1], bbox_extent[mid_idx], bbox_extent[mid_idx + 1])

    return bbox


def bbox_to_polygon_coordinates(bbox):
    """
    This function send polygon coordinates from a gives bbox

    :param bbox: input bbox
    :return: polygon coordinates
    """
    (x_min, y_min, x_max, y_max) = bbox

    return [[[x_min, y_min], [x_min, y_max], [x_max, y_max], [x_max, y_min], [x_min, y_min]]]


import decimal


def deduce_rounding_value_from_float(float_value):
    """
    From a float value return the number of digits behind the decimal point

    :param float_value: float value
    :return: number of digits behind the decimal point
    """
    d = decimal.Decimal(str(float_value))

    return abs(d.as_tuple().exponent)


def deduce_precision_from_round(rounding_value):
    """
    From the number of digits behind the decimal return the precision associate to given number

    :param rounding_value: number of digits behind the decimal point
    :return: precision associate to input rounding_value
    """

    precision_value = (1 / 10 ** rounding_value)

    return precision_value



from geoformat.conversion.feature_conversion import feature_deserialize

def field_statistics(geolayer, statistic_field, bessel_correction=True):
    """
    This function return a dictionary with statistics result in it.

        statistic_key : SUM, MEAN, MIN, MAX, RANGE, STD, COUNT, FIRST, LAST, VARIANCE, WEIGHTED_MEAN, COVARIANCE

    INPUT / OUTPUT field type table :
         One field is required :
            SUM : if original field type is real or integer       | output original field type
            MEAN : if original field type is real or integer      | output real
            MIN : if original field type is real or integer       | output original field type
            MAX : if original field type is real or integer       | output original field type
            RANGE : if original field type is real or integer     | output original field type
            STD : if original field type is real or integer       | output real
            COUNT :  original field type doesn't matter           | output integer
            FIRST : original field type doesn't matter            | output original field type
            LAST : original field type doesn't matter             | output original field type
            VARIANCE : if original field type is real or integer  | output real
            ALL: if you use this key all statistics key just above will be use


        Two fields are required :
            WEIGHTED_MEAN : fields type is real or integer        | output real
            COVARIANCE : fields type is real or integer           | output real
            ALL: if you use this key all statistics key just above will be use



    statistic_fields :
        structure example [(field_name_A, statistic_key),
                           ((field_name_a, field_name_c), statistic_key),
                           (...),
                           (field_name_X, statistic_key_1),
                           (field_name_X, statistic_key_2)]

    bassel_correction : if True in case of STD or VARIANCE calculation apply bessel correction :
        https://en.wikipedia.org/wiki/Bessel%27s_correction

    output structure : dict
        output example : {  field_name_a: {statistic_key: statistic value}
                            (field_name_a, field_name_c) : {statistic_key: statistic value}
                            field_name_x {statistic_key_1: statistic value,
                                         {statistic_key_2: statistic value}
                        }

    :param geolayer: input geolayer that contains data
    :param statistic_field:
    :param bessel_correction: if True in case of STD or VARIANCE calculation apply bessel correction
    :return: dict
    """

    valid_field_type = {'Integer', 'IntegerList', 'Real', 'RealList', 'Date', 'Time', 'DateTime'}
    statistic_keys = {'SUM', 'MEAN', 'MIN', 'MAX', 'RANGE', 'STD', 'COUNT', 'FIRST', 'LAST', 'VARIANCE',
                      'WEIGHTED_MEAN', 'COVARIANCE'}
    all_univariate_statistics_order = ['COUNT', 'SUM', 'RANGE', 'MIN', 'MEAN', 'MAX', 'STD', 'FIRST', 'LAST',
                                       'VARIANCE']
    all_bivariate_statistics_order = ['WEIGHTED_MEAN', 'COVARIANCE']

    # Check if ALL key is used and update statistic_field
    all_statistic_field = []
    for i_stat, stat in enumerate(statistic_field):
        (field_name, field_stat) = stat
        if field_stat == 'ALL':
            if isinstance(field_name, (list, tuple)):
                new_all = [[field_name, value] for value in all_bivariate_statistics_order]
            else:
                new_all = [[field_name, value] for value in all_univariate_statistics_order]
            all_statistic_field.extend(new_all)
        else:
            all_statistic_field.append(stat)
    statistic_field = all_statistic_field

    # test if field in statistic_field are valid
    for i_stat, (field_value, statistic_type) in enumerate(statistic_field):
        if isinstance(field_value, str):
            if 'fields' in geolayer['metadata']:
                field_name_type = geolayer["metadata"]['fields'][field_value]['type']
                if field_name_type not in valid_field_type or statistic_type not in statistic_keys:
                    statistic_field[i_stat] = None
        elif isinstance(field_value, (list, tuple, set)):
            for field_name in field_value:
                if isinstance(field_name, str):
                    if 'fields' in geolayer['metadata']:
                        field_name_type = geolayer["metadata"]['fields'][field_name]['type']
                        if field_name_type not in valid_field_type or statistic_type not in statistic_keys:
                            statistic_field[i_stat] = None
        else:
            print('error : field_name type is not correct')

    statistic_field = [value for value in statistic_field if value is not None]

    statistic_result = [None] * len(statistic_field)

    for fid, (i_feat, feature) in enumerate(geolayer['features'].items()):

        # if feature is serialized
        if 'feature_serialize' in geolayer['metadata']:
            if geolayer['metadata']['feature_serialize']:
                feature = feature_deserialize(feature=feature, bbox=False)

        for i, (field_name, statistic_type) in enumerate(statistic_field):

            if statistic_type.upper() in {"COVARIANCE", "WEIGHTED_MEAN"}:
                (field_name_a, field_name_b) = field_name
                value_field_a = feature['attributes'].get(field_name_a)
                value_field_b = feature['attributes'].get(field_name_b)
                feature_value = None
            else:
                feature_value = feature['attributes'].get(field_name)

            if feature_value is not None and statistic_type.upper() == 'SUM':
                if statistic_result[i] is None:
                    statistic_result[i] = 0
                statistic_result[i] += feature_value
            elif feature_value is not None and statistic_type.upper() == 'MEAN':
                if fid == 0:
                    statistic_result[i] = feature_value
                else:
                    statistic_result[i] = (statistic_result[i] * fid + feature_value) / (
                            fid + 1)
            elif statistic_type.upper() == 'WEIGHTED_MEAN':
                if value_field_a is not None and value_field_b is not None:
                    # TODO rewrite first and last in loop
                    if fid == 0:
                        statistic_result[i] = (value_field_a * value_field_b, value_field_b)
                    else:
                        statistic_result[i] = (
                            statistic_result[i][0] + value_field_a * value_field_b,
                            statistic_result[i][1] + value_field_b)
                    # last iteration
                    if fid == len(geolayer['features']) - 1:
                        statistic_result[i] = statistic_result[i][0] / statistic_result[i][1]
            elif statistic_type.upper() == 'COVARIANCE':
                if fid == 0:
                    product_sum = 0.
                    means = field_statistics(geolayer, [(field_name_a, 'MEAN'), (field_name_b, 'MEAN')])
                    mean_a = means[field_name_a]['MEAN']
                    mean_b = means[field_name_b]['MEAN']
                product = (value_field_a - mean_a) * (value_field_b - mean_b)
                product_sum += product

                # if last iteration
                if fid == len(geolayer['features']) - 1:
                    result_value = product_sum / (len(geolayer['features']))
                    statistic_result[i] = result_value

            elif statistic_type.upper() in {'MIN', 'MAX'}:
                if statistic_type.upper() == 'MAX':
                    operator = max
                else:
                    operator = min

                if feature_value is not None and statistic_result[i] is None:
                    statistic_result[i] = feature_value
                else:
                    if feature_value is not None:
                        statistic_result[i] = operator(statistic_result[i], feature['attributes'].get(field_name))

            elif statistic_type.upper() == 'RANGE':
                if fid == 0:
                    save_min = feature['attributes'][field_name]
                    save_max = feature['attributes'][field_name]
                save_min = min(save_min, feature['attributes'][field_name])
                save_max = max(save_max, feature['attributes'][field_name])
                statistic_result[i] = save_max - save_min

            elif statistic_type.upper() in ['STD', 'VARIANCE']:
                if fid == 0:
                    # saving values in a list
                    statistic_result[i] = [0] * len(geolayer['features'])
                    statistic_result[i][0] = feature['attributes'][field_name]
                else:
                    statistic_result[i][fid] = feature['attributes'][field_name]
                # if last iteration
                if fid == len(geolayer['features']) - 1:
                    nb_value = len(statistic_result[i])
                    mean_value = sum(statistic_result[i]) / nb_value
                    mean_deviation = [(value - mean_value) ** 2 for value in statistic_result[i]]

                    if bessel_correction and nb_value > 1:
                        result_value = sum(mean_deviation) / (nb_value - 1)
                    else:
                        result_value = sum(mean_deviation) / nb_value

                    if statistic_type.upper() == 'STD':
                        result_value = result_value ** 0.5

                    statistic_result[i] = result_value

            # for all field type
            if statistic_type.upper() == 'COUNT':
                if statistic_result[i] is None:
                    statistic_result[i] = 0
                statistic_result[i] += 1
            elif statistic_type.upper() == 'FIRST':
                if fid == 0:
                    statistic_result[i] = feature['attributes'][field_name]
            elif statistic_type.upper() == 'LAST':
                statistic_result[i] = feature['attributes'][field_name]

    # write results in dico result
    dico_result = {}
    for i_stat, (field_name, statistic_type) in enumerate(statistic_field):
        try:
            dico_result[field_name][statistic_type] = statistic_result[i_stat]
        except KeyError:
            dico_result[field_name] = {statistic_type: statistic_result[i_stat]}

    return dico_result
